package datavisualizer;

import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.SAXException;
import org.xml.sax.Attributes;

import javawrapper.ModelException;

/**
 * Plugs into the Xerces SAX parser to handle the incoming data when the
 * detailed output setup XML file is parsed.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 */

public class DetailedOutputFileSetupParseHandler
    extends DefaultHandler {

  /**Detailed output file manager which owns this parser.*/
  private DetailedOutputFileManager m_oManager;

  /**Set of tree settings into which to put data.*/
  private javawrapper.DetailedTreeSettings m_oTreeSettings;

  /**Set of grid settings into which to put data.*/
  private javawrapper.DetailedGridSettings m_oGridSettings;

  /**String buffer to collect data in our parser*/
  private StringBuffer m_sBuf = new StringBuffer();

  //These are flags which are set by startElement so we know what to do when
  //we get the character feed from characters.
  private boolean m_bIsFloat,
      /**<Indicates the data just parsed is a float data member.*/
      m_bIsInt, /**<Indicates the data just parsed is an int data member.*/
      m_bIsChar, /**<Indicates the data just parsed is a char data member.*/
      m_bIsBool, /**<Indicates the data just parsed is a bool data member.*/
      m_bIsPackage, /**<Indicates that these are grid package values*/
      m_bIsTimesteps,
      /**<Indicates the data just parsed is the number of timesteps.*/
      m_bIsXPlotLength,
      /**<Indicates the data just parsed is the plot X length.*/
      m_bIsYPlotLength,
      /**<Indicates the data just parsed is the plot Y length.*/
      m_bIsNumYearsPerTimestep;

  /**<Indicates the data just parsed is the number of years per timestep.*/

  /**
   * Constructor.  This is passed an object into which to put the parsed data.
   * @param oManager The DetailedOutputFileViewer object into which the data
   * goes.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public DetailedOutputFileSetupParseHandler(DetailedOutputFileManager oManager) {
    m_oManager = oManager;
    m_bIsFloat = false;
    m_bIsInt = false;
    m_bIsChar = false;
    m_bIsBool = false;
    m_bIsTimesteps = false;
    m_bIsXPlotLength = false;
    m_bIsYPlotLength = false;
    m_bIsPackage = false;

    m_oTreeSettings = null;
    m_oGridSettings = null;
  }

  /**
   * Called when the parser hits a new opening tag.  This is overridden from
   * the base class.  This function is only interested in the following tags:
   * <ul>
   * <li>ou_treeInfo - the DetailedOutputFileManager is used to create a new
   * DetailedTreeSettings object.  The object is then stashed in
   * m_oTreeSettings.</li>
   * <li>ou_float - m_bIsFloat is set to true.</li>
   * <li>ou_int - m_bIsInt is set to true.</li>
   * <li>ou_char - m_bIsChar is set to true.</li>
   * <li>ou_bool - m_bIsBool is set to true.</li>
   * <li>tr_species - the species name is passed to m_oManager.</li>
   * <li>timesteps - m_bIsTimesteps is set to true.</li>
   * <li>plot_lenX - m_bIsXPlotLength is set to true.</li>
   * <li>plot_lenY - m_bIsYPlotLength is set to true.</li>
   * <li>ou_gridInfo - the DetailedOutputFileManager is used to create a new
   * DetailedGridSettings object.  The object is then stashed in
   * m_oGridSettings.</li>
   * <li>Something starting with "ma_package" - m_bIsPackage is set to
   * true.</li>
   * </ul>
   * @param uri the Namespace URI (ignored)
   * @param localName the local name (what this function looks at)
   * @param qName the qualified (prefixed) name (ignored)
   * @param attributes The tag's attributes
   * @throws SAXException if there are any problems.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>March 21, 2005:  Fixed a bug so that package members could be
   * differentiated (LEM)
   */
  public void startElement(java.lang.String uri,
                           java.lang.String localName,
                           java.lang.String qName,
                           Attributes attributes) throws SAXException {
    try {

      //Initialize our string buffer if necessary
      if (m_sBuf.length() > 0) {
        m_sBuf = null;
        m_sBuf = new StringBuffer();
      }

      //Do the appropriate thing, depending on the tag name
      if (qName.equals("ou_float")) {

        //Set the float codes flag to true
        m_bIsFloat = true;

      }
      else if (qName.equals("ou_int")) {

        //Set the int codes flag to true
        m_bIsInt = true;

      }
      else if (qName.equals("ou_char")) {

        //Set the char codes flag to true
        m_bIsChar = true;

      }
      else if (qName.equals("ou_bool")) {

        //Set the bool codes flag to true
        m_bIsBool = true;

      }
      else if (qName.startsWith("ma_package")) { //LEM 03-25-05
        m_bIsPackage = true;
      }
      else if (qName.equals("ou_treeInfo")) {

        //Null out the grid settings so we know that anything we're getting
        //now is for trees
        m_oGridSettings = null;

        //Get the type number and species name
        String sType = attributes.getValue("type");
        String sSpecies = attributes.getValue("species");

        //Transform sType into an int
        if (sType.equalsIgnoreCase("seed")) {
          m_oTreeSettings = m_oManager.CreateNewTreeSettings(sSpecies,
              javawrapper.TreePopulation.SEED);
        }
        else if (sType.equalsIgnoreCase("seedling")) {
          m_oTreeSettings = m_oManager.CreateNewTreeSettings(sSpecies,
              javawrapper.TreePopulation.SEEDLING);
        }
        else if (sType.equalsIgnoreCase("sapling")) {
          m_oTreeSettings = m_oManager.CreateNewTreeSettings(sSpecies,
              javawrapper.TreePopulation.SAPLING);
        }
        else if (sType.equalsIgnoreCase("adult")) {
          m_oTreeSettings = m_oManager.CreateNewTreeSettings(sSpecies,
              javawrapper.TreePopulation.ADULT);
        }
        else if (sType.equalsIgnoreCase("stump")) {
          m_oTreeSettings = m_oManager.CreateNewTreeSettings(sSpecies,
              javawrapper.TreePopulation.STUMP);
        }
        else if (sType.equalsIgnoreCase("snag")) {
          m_oTreeSettings = m_oManager.CreateNewTreeSettings(sSpecies,
              javawrapper.TreePopulation.SNAG);
        }
      }
      else if (qName.equals("ou_gridInfo")) {

        //Null out the tree settings so we know that anything we're getting
        //now is for trees
        m_oTreeSettings = null;

        //Get the grid name
        String sGridName = attributes.getValue("gridName");

        //Create the new grid settings object
        m_oGridSettings = m_oManager.CreateNewGridSettings(sGridName);

      }
      else if (qName.equals("tr_species")) {

        //Extract the species name from the attributes
        String sName = attributes.getValue("speciesName");

        //Convert underscores to spaces
        //sName = sName.replace('_', ' ');

        m_oManager.AddSpeciesName(sName);
      }
      else if (qName.equals("timesteps")) {
        m_bIsTimesteps = true;
      }
      else if (qName.equals("plot_lenX")) {
        m_bIsXPlotLength = true;
      }
      else if (qName.equals("plot_lenY")) {
        m_bIsYPlotLength = true;
      }
      else if (qName.equals("yearsPerTimestep")) {
        m_bIsNumYearsPerTimestep = true;
      }

      //Any other tags - we don't care.
    }
    catch (ModelException oErr) {
      //Unfortunately, we have to wrap our own exceptions as SAXExceptions.
      SAXException e = new SAXException(oErr.sMessage);
      throw (e);
    }
  }

  /**
   * Reads character data from the XML file.  Whether or not the data is
   * ignored, and what is done with it if it is not ignored, depends on flags
   * that have been set.
   * <p>If m_bIsFloat is true, AddNewFloat() is called for
   * m_oTreeSettings.  If mp_oTreeMapSettings is null, an error is thrown.
   * <p>If m_bIsInt is true, AddNewInt() is called for
   * m_oTreeSettings.  If mp_oTreeMapSettings is null, an error is thrown.
   * <p>If m_bIsChar is true, AddNewChar() is called for
   * m_oTreeSettings.  If mp_oTreeMapSettings is null, an error is thrown.
   * <p>If m_bIsBool is true, AddNewFloat() is called for
   * m_oTreeSettings.  If mp_oTreeMapSettings is null, an error is thrown.
   * <p>If m_bIsTimesteps is true, the number of timesteps is sent to
   * m_oManager.
   * <p>After execution any flag turned on is turned back off.
   * @param ch The characters from the XML document.
   * @param start - The start position in the array.
   * @param length - The number of characters to read from the array.
   * @throws SAXException if any of the described cases above is true.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void characters(char[] ch,
                         int start,
                         int length) throws SAXException {

    //Trim the string of both characters that don't matter to XML and of
    //spaces
    String sNewData = javawrapper.SaxParseTools.XMLTrim(new String(ch, start,
        length));
    //If there is already data in the buffer, append the new string
    if (m_sBuf.length() > 0) {
      m_sBuf.append(sNewData);
    }
    else {

      //Try trimming the string of spaces - if there's nothing left, exit;
      //otherwise, append the untrimmed string
      if (sNewData.trim().length() > 0) {
        m_sBuf.append(sNewData);
      }
    }

  }

  /**
   * Called at the end of an XML tag.  Whether or not the data is
   * ignored, and what is done with it if it is not ignored, depends on flags
   * that have been set.
   * <p>If m_bIsFloat is true, AddNewFloat() is called for
   * m_oTreeSettings.  If mp_oTreeMapSettings is null, an error is thrown.
   * <p>If m_bIsInt is true, AddNewInt() is called for
   * m_oTreeSettings.  If mp_oTreeMapSettings is null, an error is thrown.
   * <p>If m_bIsChar is true, AddNewChar() is called for
   * m_oTreeSettings.  If mp_oTreeMapSettings is null, an error is thrown.
   * <p>If m_bIsBool is true, AddNewFloat() is called for
   * m_oTreeSettings.  If mp_oTreeMapSettings is null, an error is thrown.
   * <p>If m_bIsTimesteps is true, the number of timesteps is sent to
   * m_oManager.
   * <p>After execution any flag turned on is turned back off.
   * @param sURI the Namespace URI (ignored)
   * @param sLocalName the local name (what this function looks at)
   * @param sQName the qualified (prefixed) name (ignored)
   * @throws SAXException if there were problems assigning the data.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>March 21, 2005:  Fixed a bug so that package members could be
   * differentiated (LEM)
   */
  public void endElement(java.lang.String sURI,
                         java.lang.String sLocalName,
                         java.lang.String sQName) throws SAXException {

    //Make sure there's data in the string buffer
    if (m_sBuf.length() == 0) return;

    //Check our flags to see what we want to do
    if (m_bIsFloat) {

      if (null != m_oTreeSettings) {

        //This is a float code from a tree setting - add it
        m_oTreeSettings.AddFloat(m_sBuf.toString());

      }
      else {

        //This is a float code from a grid setting - add it if not a package
        if (m_bIsPackage == false) { //LEM 03-25-05
          m_oGridSettings.AddFloat(m_sBuf.toString());
        }

      }

      //Reset our flag
      m_bIsFloat = false;

    }
    else if (m_bIsInt) {

      if (null != m_oTreeSettings) {

        //This is a int code from a tree setting - add it
        m_oTreeSettings.AddInt(m_sBuf.toString());

      }
      else {

        //This is a int code from a grid setting - add it if not a package
        if (m_bIsPackage == false) { //LEM 03-25-05
          m_oGridSettings.AddInt(m_sBuf.toString());
        }

      }

      //Reset our flag
      m_bIsInt = false;

    }
    else if (m_bIsChar) {

      if (null != m_oTreeSettings) {

        //This is a char code from a tree setting - add it
        m_oTreeSettings.AddChar(m_sBuf.toString());

      }
      else {

        //This is a char code from a grid setting - add it if not a package
        if (m_bIsPackage == false) { //LEM 03-25-05
          m_oGridSettings.AddChar(m_sBuf.toString());
        }

      }

      //Reset our flag
      m_bIsChar = false;

    }
    else if (m_bIsBool) {

      if (null != m_oTreeSettings) {

        //This is a bool code from a tree setting - add it
        m_oTreeSettings.AddBool(m_sBuf.toString());

      }
      else {

        //This is a bool code from a grid setting - add it if not a package
        if (m_bIsPackage == false) { //LEM 03-25-05
          m_oGridSettings.AddBool(m_sBuf.toString());
        }

      }

      //Reset our flag
      m_bIsBool = false;

    }
    else if (m_bIsTimesteps) {

      int iTimesteps = new Integer(m_sBuf.toString()).intValue();
      m_oManager.SetParFileTimesteps(iTimesteps);

      m_bIsTimesteps = false;
    }
    else if (m_bIsXPlotLength) {

      float f = new Float(m_sBuf.toString()).floatValue();
      m_oManager.SetXPlotLength(f);

      m_bIsXPlotLength = false;
    }
    else if (m_bIsYPlotLength) {

      float f = new Float(m_sBuf.toString()).floatValue();
      m_oManager.SetYPlotLength(f);

      m_bIsYPlotLength = false;
    }
    else if (m_bIsNumYearsPerTimestep) {

      float f = new Float(m_sBuf.toString()).floatValue();
      m_oManager.SetNumberOfYearsPerTimestep(f);

      m_bIsNumYearsPerTimestep = false;
    }

    //If this is the end of a set of package values, reset the flag
    //LEM 03-25-05
    if (sQName.startsWith("ma_package")) {
      m_bIsPackage = false;
    }
  }

}
