package javawrapper;

import javax.swing.*;
import javax.swing.tree.*;
import javax.help.HelpBroker;
import java.awt.*;
import java.awt.event.*;
import java.util.Vector;
import javax.help.HelpBroker;
import javax.help.CSH;

/**
 * Allows users to make changes to the model flow and behavior order.
 * <p>Copyright: Copyright (c) 2003 Charles D. Canham</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 */

public class ModelFlowSetup
    extends JDialog
    implements ActionListener,
    javax.swing.event.TreeSelectionListener {

  /**GUIManager object.*/
  protected GUIManager m_oManager;

  /**Tree showing flow*/
  protected JTree m_jTree = null;

  /**Data model of tree showing flow*/
  protected DefaultTreeModel m_jTreeModel;

  /**TreePopulation object*/
  protected TreePopulation m_oPop;

  protected String[]
      /**List of behavior groups*/
      mp_sBehaviorGroupNames = null,
      /**List of grid names*/
      mp_sGridNames = null;

  /**Our copy of the Behaviors.  The first index is behavior group
   * number; the second is a list of all Behaviors for that group, in order.
   * These are clones, so we can mess with them here, and then load the data
   * back when we're done.*/
  protected Behavior[][] mp_oBehaviors;

  /**Max dimensions of the main program flow dialog*/
  protected Dimension m_oMaxDimension;

  private String m_sHelpID = "windows.model_flow_window";

  protected final int
      /**Flag indicating that the tree should display behaviors at the
       * highest level*/
      BEHAVIOR_FIRST = 0,
      /**Flag indicating that the tree should display data at the highest
       * level, and for trees, type at a higher level than species*/
      TYPE_FIRST = 1,
      /**Flag indicating that the tree should display data at the highest
       * level, and for trees, species at a higher level than type*/
      SPECIES_FIRST = 2;

  /**Mode by which the tree should be built - will be one of the finals
   * below*/
  protected int m_iMode = TYPE_FIRST;

  /**
   * Constructor.
   * @param oParent Window which is the parent of this dialog.
   * @param oManager GUIManager object.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public ModelFlowSetup(JFrame oParent, GUIManager oManager) {
    super(oParent, "Edit Simulation Flow", true);
    m_oManager = oManager;
    m_oPop = oManager.GetTreePopulation();
    m_oMaxDimension = new Dimension(oParent.getWidth() - 50,
                                    oParent.getHeight() - 50);

    //Help ID
    oManager.GetHelpBroker().enableHelpKey(this.getRootPane(), m_sHelpID, null);

    try {

      //Load the existing set of behaviors
      BehaviorTypeBase[] p_oBehaviorGroupList = m_oManager.GetAllBehaviorGroups();
      Behavior[] p_oBehaviorList;
      int i, j;

      //Declare the behavior group list
      mp_sBehaviorGroupNames = new String[p_oBehaviorGroupList.length];
      mp_oBehaviors = new Behavior[p_oBehaviorGroupList.length][];

      //Go through all the behaviors one at a time
      for (i = 0; i < p_oBehaviorGroupList.length; i++) {

        //Grab the name of this behavior group
        mp_sBehaviorGroupNames[i] = p_oBehaviorGroupList[i].GetName();

        p_oBehaviorList = p_oBehaviorGroupList[i].GetAllBehaviors();

        if (p_oBehaviorList != null) {

          //Clone this list
          mp_oBehaviors[i] = new Behavior[p_oBehaviorList.length];
          for (j = 0; j < p_oBehaviorList.length; j++) {
            mp_oBehaviors[i][j] = (Behavior) p_oBehaviorList[j].clone();
          }
        }
        else {
          //Declare this to have 0 length - don't want it to be null
          mp_oBehaviors[i] = new Behavior[0];
        }
      }

      //Retrieve the list of all grid objects
      Vector oGrids = new Vector(0);
      Grid[] oSetGrids;
      Grid oGrid;
      if (null != p_oBehaviorGroupList) {
        for (i = 0; i < p_oBehaviorGroupList.length; i++) {
          oSetGrids = p_oBehaviorGroupList[i].GetAllGridObjects();
          if (null != oSetGrids) {
            for (j = 0; j < oSetGrids.length; j++) {
              oGrids.add(oSetGrids[j]);
            }
          }
        }
      }

      mp_sGridNames = new String[oGrids.size()];
      for (i = 0; i < mp_sGridNames.length; i++) {
        oGrid = (Grid) oGrids.elementAt(i);
        mp_sGridNames[i] = oGrid.GetName();
      }

      //Create the GUI
      CreateGUI();
    }
    catch (ModelException oErr) {
      ErrorGUI oHandler = new ErrorGUI(this);
      oHandler.WriteErrorMessage(oErr);
    }
  }

  /**
   * Sets the behavior list for a species/type combo in the temporary list
   * held in this object.  This does NOT make the changes in the official
   * list held in the GUIManager object.  The GUI is updated to reflect
   * changes.
   * @param iSpecies Species to set for.
   * @param iType Type to set for.
   * @param p_sBehaviors List of behavior display names.
   * @throws ModelException passed through from called methods.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected void SetBehaviorList(int iSpecies, int iType, String[] p_sBehaviors) throws
      ModelException {

    TreePopulation oPop = m_oManager.GetTreePopulation();
    SpeciesTypeCombo oCombo,
        oNewToSet = new SpeciesTypeCombo(iSpecies, iType, oPop);
    int iBehIndex, i, j, k;

    //First remove all existing assignments for this type/species
    for (i = 0; i < mp_oBehaviors.length; i++) {
      for (j = 0; j < mp_oBehaviors[i].length; j++) {
        for (k = 0; k < mp_oBehaviors[i][j].GetNumberOfCombos(); k++) {
          oCombo = mp_oBehaviors[i][j].GetSpeciesTypeCombo(k);
          if (oCombo.equals(oNewToSet)) {
            mp_oBehaviors[i][j].DeleteSpeciesTypeCombo(k);
          }
        }
      }
    }

    //Now match all behaviors and assign them the species/type combo
    if (p_sBehaviors == null || p_sBehaviors.length == 0) {
      return;
    }
    for (iBehIndex = 0; iBehIndex < p_sBehaviors.length; iBehIndex++) {

      //Find this behavior
      for (i = 0; i < mp_oBehaviors.length; i++) {
        for (j = 0; j < mp_oBehaviors[i].length; j++) {
          if (mp_oBehaviors[i][j].GetDescriptor().equals(p_sBehaviors[iBehIndex])) {
            mp_oBehaviors[i][j].AddSpeciesTypeCombo(oNewToSet);
          }
        }
      }
    }

    BuildTree();
  }

  /**
   * Draws the window displaying the data as a tree.
   * @throws ModelException passed through from called methods.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected void CreateGUI() throws ModelException {
    JPanel jContentPanel = new JPanel(); //holds everything

    BuildTree();

    m_jTree.getSelectionModel().setSelectionMode
        (TreeSelectionModel.SINGLE_TREE_SELECTION);

    //Listen for when the selection changes.
    m_jTree.addTreeSelectionListener(this);

    m_jTree.setFont(new ModelFont());
    JScrollPane jScroller = new JScrollPane(m_jTree);
    jScroller.setPreferredSize(new Dimension(
        (int) java.lang.Math.min(m_oMaxDimension.getWidth(), 400),
        (int) java.lang.Math.min(m_oMaxDimension.getHeight(), 400)));

    //Panel for switch buttons
    JButton jBehaviorView = new JButton("View by behavior");
    jBehaviorView.setFont(new ModelFont());
    jBehaviorView.setActionCommand("BehaviorFirstMode");
    jBehaviorView.addActionListener(this);

    JButton jSpeciesView = new JButton("View by data - species first");
    jSpeciesView.setFont(new ModelFont());
    jSpeciesView.setActionCommand("SpeciesFirstMode");
    jSpeciesView.addActionListener(this);

    JButton jTypeView = new JButton("View by data - type first");
    jTypeView.setFont(new ModelFont());
    jTypeView.setActionCommand("TypeFirstMode");
    jTypeView.addActionListener(this);

    JPanel jViewButtonPanel = new JPanel();
    FlowLayout jButtonLayout = new FlowLayout();
    jButtonLayout.setAlignment(FlowLayout.RIGHT);
    jViewButtonPanel.add(jBehaviorView);
    jViewButtonPanel.add(jSpeciesView);
    jViewButtonPanel.add(jTypeView);

    jContentPanel.setLayout(new BorderLayout());
    jContentPanel.add(jScroller, BorderLayout.CENTER);
    jContentPanel.add(new OKCancelButtonPanel(this, m_oManager.GetHelpBroker(),
                                              m_sHelpID), BorderLayout.PAGE_END);
    jContentPanel.add(jViewButtonPanel, BorderLayout.PAGE_START);

    setContentPane(jContentPanel);
  }

  /**
   * Gets a vector of behavior names for a type/species combo from our
   * temporary behavior copy set.
   * @param iSpecies Tree species.
   * @param iType Tree type.
   * @return Vector of behavior names.
   * @throws ModelException passed through from called methods.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected Vector GetBehaviorList(int iSpecies, int iType) throws
      ModelException {
    SpeciesTypeCombo oCombo;
    Vector oList = new Vector(0);
    int i, j, k;

    for (i = 0; i < mp_oBehaviors.length; i++) {
      for (j = 0; j < mp_oBehaviors[i].length; j++) {
        if (mp_oBehaviors[i][j].m_bIsEnabled) {
          for (k = 0; k < mp_oBehaviors[i][j].GetNumberOfCombos(); k++) {
            oCombo = mp_oBehaviors[i][j].GetSpeciesTypeCombo(k);
            if (oCombo.GetSpecies() == iSpecies &&
                oCombo.GetType() == iType) {
              oList.add(mp_oBehaviors[i][j].GetDescriptor());
              break;
            }
          }
        }
      }
    }
    return oList;
  }

  /**
   * Gets the enabled behaviors that apply to a grid from our temporary
   * behavior copy set.
   * @param sGridName Name of the grid.
   * @return Vector of behavior names.
   * @throws ModelException passed through from called methods.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected Vector GetBehaviorList(String sGridName) throws ModelException {
    Vector oList = new Vector(0);
    Grid oGrid;
    int i, j, k;

    for (i = 0; i < mp_oBehaviors.length; i++) {
      for (j = 0; j < mp_oBehaviors[i].length; j++) {
        if (mp_oBehaviors[i][j].m_bIsEnabled) {
          for (k = 0; k < mp_oBehaviors[i][j].GetNumberOfGrids(); k++) {
            oGrid = mp_oBehaviors[i][j].GetGrid(k);
            if (oGrid.GetName().equals(sGridName)) {
              oList.add(mp_oBehaviors[i][j].GetDescriptor());
              break;
            }
          }
        }
      }
    }
    return oList;
  }

  /**
   * Builds the tree displayed in the window with data first; for trees, the
   * branch levels are type, then species, then behavior list.
   * @param jRoot Node upon which the tree is built.
   * @throws ModelException passed through from called methods.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected void BuildTypeFirstTree(DefaultMutableTreeNode jRoot) throws
      ModelException {
    Vector[][] p_oBehaviorsByCombo =
        new Vector[m_oPop.GetNumberOfTypes()][m_oPop.GetNumberOfSpecies()];

    int i, j, k;

    //Build the list of behaviors for each species/type combo
    for (i = 0; i < p_oBehaviorsByCombo.length; i++) {
      for (j = 0; j < p_oBehaviorsByCombo[i].length; j++) {
        p_oBehaviorsByCombo[i][j] = new Vector();
        Vector p_oBehaviors = GetBehaviorList(j, i);
        if (p_oBehaviors != null) {
          for (k = 0; k < p_oBehaviors.size(); k++) {
            p_oBehaviorsByCombo[i][j].add(p_oBehaviors.elementAt(k));
          }
        }
      }
    }

    //Add the grid node
    jRoot.add(BuildGridNode());

    //Create a home node for all trees
    DefaultMutableTreeNode jTreeRoot = new DefaultMutableTreeNode(
        new TreeNodeInfo("Trees", TreeNodeInfo.IGNORE));

    //Create a tree node for each life history stage
    for (i = 0; i < p_oBehaviorsByCombo.length; i++) {

      DefaultMutableTreeNode jLifeHistoryStage = new
          DefaultMutableTreeNode(new TreeNodeInfo(m_oPop.GetTypeNameFromCode(i),
                                                  TreeNodeInfo.TYPE));
      jTreeRoot.add(jLifeHistoryStage);

      for (j = 0; j < p_oBehaviorsByCombo[i].length; j++) {

        //Create a tree node for each species
        DefaultMutableTreeNode jSpecies = new DefaultMutableTreeNode(
            new TreeNodeInfo(m_oPop.GetSpeciesNameFromCode(j).replace('_', ' '),
                             TreeNodeInfo.SPECIES));
        jLifeHistoryStage.add(jSpecies);

        //Now add a node for each behavior
        for (k = 0; k < p_oBehaviorsByCombo[i][j].size(); k++) {

          DefaultMutableTreeNode jBehavior = new DefaultMutableTreeNode(
              new TreeNodeInfo( (String) p_oBehaviorsByCombo[i][j].elementAt(k),
                               TreeNodeInfo.BEHAVIOR));

          jSpecies.add(jBehavior);
        }

        //If there are no nodes under this species, add a node saying no
        //Behaviors
        if (jSpecies.getChildCount() == 0) {
          jSpecies.add(new DefaultMutableTreeNode(
              new TreeNodeInfo("No behaviors", TreeNodeInfo.BEHAVIOR)));
        }
      }
    }
    jRoot.add(jTreeRoot);
  }

  /**
   * Builds the tree displayed in the window with data first; for trees, the
   * branch levels are species, then type, then behavior list.
   * @param jRoot Node upon which the tree is built.
   * @throws ModelException passed through from called methods.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected void BuildSpeciesFirstTree(DefaultMutableTreeNode jRoot) throws
      ModelException {
    Vector[][] p_oBehaviorsByCombo =
        new Vector[m_oPop.GetNumberOfSpecies()][m_oPop.GetNumberOfTypes()];

    int i, j, k;

    //Build the list of behaviors for each species/type combo
    for (i = 0; i < p_oBehaviorsByCombo.length; i++) {
      for (j = 0; j < p_oBehaviorsByCombo[i].length; j++) {
        p_oBehaviorsByCombo[i][j] = new Vector();
        Vector p_oBehaviors = GetBehaviorList(i, j);
        if (p_oBehaviors != null) {
          for (k = 0; k < p_oBehaviors.size(); k++) {
            p_oBehaviorsByCombo[i][j].add(p_oBehaviors.elementAt(k));
          }
        }
      }
    }

    //Add the grid node
    jRoot.add(BuildGridNode());

    //Create a home node for all trees
    DefaultMutableTreeNode jTreeRoot = new DefaultMutableTreeNode(
        new TreeNodeInfo("Trees", TreeNodeInfo.IGNORE));

    //Create a tree node for each species
    for (i = 0; i < p_oBehaviorsByCombo.length; i++) {

      DefaultMutableTreeNode jSpecies = new
          DefaultMutableTreeNode(new TreeNodeInfo(m_oPop.GetSpeciesNameFromCode(
          i).replace('_', ' '), TreeNodeInfo.SPECIES));
      jTreeRoot.add(jSpecies);

      for (j = 0; j < p_oBehaviorsByCombo[i].length; j++) {

        //Create a tree node for each life history stage
        DefaultMutableTreeNode jLifeHistoryStage = new DefaultMutableTreeNode(
            new TreeNodeInfo(m_oPop.GetTypeNameFromCode(j), TreeNodeInfo.TYPE));
        jSpecies.add(jLifeHistoryStage);

        //Now add a node for each behavior
        for (k = 0; k < p_oBehaviorsByCombo[i][j].size(); k++) {

          DefaultMutableTreeNode jBehavior = new DefaultMutableTreeNode(
              new TreeNodeInfo( (String) p_oBehaviorsByCombo[i][j].elementAt(k),
                               TreeNodeInfo.BEHAVIOR));

          jLifeHistoryStage.add(jBehavior);
        }

        //If there are no nodes under this species, add a node saying no
        //Behaviors
        if (jLifeHistoryStage.getChildCount() == 0) {
          jLifeHistoryStage.add(new DefaultMutableTreeNode(
              new TreeNodeInfo("No behaviors", TreeNodeInfo.BEHAVIOR)));
        }
      }
    }
    jRoot.add(jTreeRoot);
  }

  /**
   * Builds a tree where the first level is behavior; the second is data,
   * with trees being type first, then species.
   * @param jRoot Node upon which to build the tree.
   * @throws ModelException passed through from called methods.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected void BuildBehaviorFirstTree(DefaultMutableTreeNode jRoot) throws
      ModelException {

    //This array determines which combos a behavior has, sorted by type
    boolean[][] p_oWhichCombos = new boolean[m_oPop.GetNumberOfTypes()][m_oPop.
        GetNumberOfSpecies()];
    SpeciesTypeCombo oCombo;
    int i, j, k, m;

    //Go through each behavior
    for (i = 0; i < mp_oBehaviors.length; i++) {
      for (j = 0; j < mp_oBehaviors[i].length; j++) {
        if (mp_oBehaviors[i][j].m_bIsEnabled) {

          DefaultMutableTreeNode jBehaviorNode = new DefaultMutableTreeNode(new
              TreeNodeInfo(mp_oBehaviors[i][j].GetDescriptor(),
                           TreeNodeInfo.BEHAVIOR));

          //Add a master grid node
          DefaultMutableTreeNode jGridRoot = new DefaultMutableTreeNode(
              new TreeNodeInfo("Grids", TreeNodeInfo.IGNORE));

          if (mp_oBehaviors[i][j].GetNumberOfGrids() == 0) {
            jGridRoot.add(new DefaultMutableTreeNode(new TreeNodeInfo(
                "No grids", TreeNodeInfo.IGNORE)));
          }
          else {
            //Add a node for each grid
            for (k = 0; k < mp_oBehaviors[i][j].GetNumberOfGrids(); k++) {
              jGridRoot.add(new DefaultMutableTreeNode(new TreeNodeInfo(
                  mp_oBehaviors[i][j].GetGrid(k).GetName(), TreeNodeInfo.GRID)));
            }
          }
          jBehaviorNode.add(jGridRoot);

          //Add a master tree node
          DefaultMutableTreeNode jTreeRoot = new DefaultMutableTreeNode(
              new TreeNodeInfo("Trees", TreeNodeInfo.IGNORE));

          //Reset the combos array
          for (k = 0; k < p_oWhichCombos.length; k++) {
            for (m = 0; m < p_oWhichCombos[k].length; m++) {
              p_oWhichCombos[k][m] = false;
            }
          }

          //Set which combos this behavior has
          for (k = 0; k < mp_oBehaviors[i][j].GetNumberOfCombos(); k++) {
            oCombo = mp_oBehaviors[i][j].GetSpeciesTypeCombo(k);
            p_oWhichCombos[oCombo.GetType()][oCombo.GetSpecies()] = true;
          }

          if (mp_oBehaviors[i][j].GetNumberOfCombos() == 0) {
            jTreeRoot.add(new DefaultMutableTreeNode(new TreeNodeInfo(
                "No trees", TreeNodeInfo.IGNORE)));
          }
          else {

            for (k = 0; k < p_oWhichCombos.length; k++) {

              DefaultMutableTreeNode jLifeHistoryStage = new
                  DefaultMutableTreeNode(new TreeNodeInfo(m_oPop.
                  GetTypeNameFromCode(k), TreeNodeInfo.TYPE));

              for (m = 0; m < p_oWhichCombos[k].length; m++) {
                if (p_oWhichCombos[k][m]) {
                  jLifeHistoryStage.add(new DefaultMutableTreeNode(new
                      TreeNodeInfo(m_oPop.GetSpeciesNameFromCode(m).replace('_',
                      ' '), TreeNodeInfo.SPECIES)));
                }
              }

              if (jLifeHistoryStage.getChildCount() > 0) {
                jTreeRoot.add(jLifeHistoryStage);
              }
            }
          }

          jBehaviorNode.add(jTreeRoot);

          jRoot.add(jBehaviorNode);
        }
      }
    }

    //If there are no behaviors defined - add a node saying so
    if (jRoot.getChildCount() == 0) {
      jRoot.add(new DefaultMutableTreeNode(new TreeNodeInfo(
          "Click here to add behaviors", TreeNodeInfo.BEHAVIOR)));
    }

  }

  /**
   * Builds a node for Grid data.  First level is grid name, second level is
   * behavior.
   * @return Root grid node.
   * @throws ModelException passed through from called methods.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected DefaultMutableTreeNode BuildGridNode() throws ModelException {
    int i, j;
    //Create a grids node
    DefaultMutableTreeNode jGridRoot = new DefaultMutableTreeNode(
        new TreeNodeInfo("Grids", TreeNodeInfo.IGNORE));
    //Build the grids tree
    for (i = 0; i < mp_sGridNames.length; i++) {
      DefaultMutableTreeNode jGrid = new DefaultMutableTreeNode(
          new TreeNodeInfo(mp_sGridNames[i], TreeNodeInfo.GRID));
      Vector p_oBehaviorsForGrid = GetBehaviorList(mp_sGridNames[i]);
      if (p_oBehaviorsForGrid.size() == 0) {
        jGrid.add(new DefaultMutableTreeNode(new TreeNodeInfo(
            "No behaviors", TreeNodeInfo.BEHAVIOR)));
      }
      else {
        for (j = 0; j < p_oBehaviorsForGrid.size(); j++) {
          jGrid.add(new DefaultMutableTreeNode(new TreeNodeInfo(
              (String) p_oBehaviorsForGrid.elementAt(j),
              TreeNodeInfo.BEHAVIOR)));
        }
      }
      jGridRoot.add(jGrid);
    }

    return jGridRoot;
  }

  /**
   * Builds the tree displayed in the window.  If it already exists, it will
   * be torn down and recreated.
   * @throws ModelException passed through from called methods.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected void BuildTree() throws ModelException {

    DefaultMutableTreeNode jRoot;

    if (m_jTree == null) {
      //Create a tree root node.
      jRoot =
          new DefaultMutableTreeNode(new TreeNodeInfo("Click a name to edit",
          TreeNodeInfo.IGNORE));
    }
    else {

      //Get the existing root and tear it down
      jRoot = (DefaultMutableTreeNode) m_jTree.getModel().getRoot();
      jRoot.removeAllChildren();
    }

    //Build the appropriate tree according to the tree mode flag
    if (m_iMode == TYPE_FIRST) {
      BuildTypeFirstTree(jRoot);
    }
    else if (m_iMode == BEHAVIOR_FIRST) {
      BuildBehaviorFirstTree(jRoot);
    }
    else if (m_iMode == SPECIES_FIRST) {
      BuildSpeciesFirstTree(jRoot);
    }

    //Create our tree, put it in a scroll pane, and display it
    if (m_jTree == null) {
      m_jTreeModel = new DefaultTreeModel(jRoot);
      m_jTree = new JTree(m_jTreeModel);
    }
    else {
      m_jTreeModel.setRoot(jRoot);
      m_jTreeModel.reload();
    }
  }

  /**
   * Controls actions for this window.
   * @param e ActionEvent.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void actionPerformed(ActionEvent e) {
    if (e.getActionCommand().equals("Cancel")) {
      this.setVisible(false);
      this.dispose();
    }
    else if (e.getActionCommand().equals("OK")) {
      try {
        AssignData();
        this.setVisible(false);
        this.dispose();
      }
      catch (ModelException oErr) {
        //Display the error but don't close the window
        ErrorGUI oHandler = new ErrorGUI(this);
        oHandler.WriteErrorMessage(oErr);
      }
    }
    else if (e.getActionCommand().equals("BehaviorFirstMode")) {
      try {
        //Requested switch to behavior first mode in the tree view

        //If we're already there, exit
        if (m_iMode == BEHAVIOR_FIRST) {
          return;
        }

        m_iMode = BEHAVIOR_FIRST;
        BuildTree();
      }
      catch (ModelException oErr) {
        ErrorGUI oHandler = new ErrorGUI(this);
        oHandler.WriteErrorMessage(oErr);
      }
    }
    else if (e.getActionCommand().equals("TypeFirstMode")) {
      try {
        //Requested switch to type first mode in the tree view

        //If we're already there, exit
        if (m_iMode == TYPE_FIRST) {
          return;
        }

        m_iMode = TYPE_FIRST;
        BuildTree();
      }
      catch (ModelException oErr) {
        ErrorGUI oHandler = new ErrorGUI(this);
        oHandler.WriteErrorMessage(oErr);
      }
    }
    else if (e.getActionCommand().equals("SpeciesFirstMode")) {
      try {
        //Requested switch to type first mode in the tree view

        //If we're already there, exit
        if (m_iMode == SPECIES_FIRST) {
          return;
        }

        m_iMode = SPECIES_FIRST;
        BuildTree();
      }
      catch (ModelException oErr) {
        ErrorGUI oHandler = new ErrorGUI(this);
        oHandler.WriteErrorMessage(oErr);
      }
    }
  }

  /**
   * Assigns the changes made in this dialog back to the behavior groups.
   * First the behavior settings are validated, and then they are copied
   * back in order to the behavior groups.
   * @throws ModelException if the settings are invalid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected void AssignData() throws ModelException {
    //Validate all behaviors
    int i, j;
    for (i = 0; i < mp_oBehaviors.length; i++) {
      for (j = 0; j < mp_oBehaviors[i].length; j++) {
        mp_oBehaviors[i][j].Validate();
      }
    }

    //Copy back the behaviors to each behavior group
    BehaviorTypeBase[] p_oRealBehaviorGroups = m_oManager.GetAllBehaviorGroups();

    for (i = 0; i < p_oRealBehaviorGroups.length; i++) {

      //Just as a check - make sure that the group names match
      if (!p_oRealBehaviorGroups[i].GetName().equals(mp_sBehaviorGroupNames[i])) {
        ModelException oErr = new ModelException();
        oErr.iErrorCode = ErrorGUI.BAD_DATA;
        oErr.sCallingFunction = "ProgramFlowSetup.AssignData - JAVA";
        oErr.sMessage = "Behavior groups do not match when assigning data.";
        throw (oErr);
      }

      p_oRealBehaviorGroups[i].mp_oChildBehaviors = null;
      p_oRealBehaviorGroups[i].mp_oChildBehaviors = new Behavior[mp_oBehaviors[
          i].length];
      for (j = 0; j < mp_oBehaviors[i].length; j++) {
        p_oRealBehaviorGroups[i].mp_oChildBehaviors[j] =
            (Behavior) mp_oBehaviors[i][j].clone();
      }
    }
  }

  /**
   * Detects user clicks on the tree.
   * @param e TreeSelectionEvent
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void valueChanged(javax.swing.event.TreeSelectionEvent e) {
    DefaultMutableTreeNode jNode = (DefaultMutableTreeNode)
        m_jTree.getLastSelectedPathComponent();
    try {
      if (jNode == null) {
        return;
      }

      TreeNodeInfo oInfo = (TreeNodeInfo) jNode.getUserObject();
      if (oInfo.GetNodeType() == TreeNodeInfo.IGNORE) {
        return;
      }

      if (oInfo.GetNodeType() == TreeNodeInfo.SPECIES) {

        //If the parent node was type, show the combo editor
        DefaultMutableTreeNode jParent = (DefaultMutableTreeNode) jNode.
            getParent();
        TreeNodeInfo oParentInfo = (TreeNodeInfo) jParent.getUserObject();
        if (oParentInfo.GetNodeType() == TreeNodeInfo.TYPE) {

          DisplayComboEdit oEditor = new DisplayComboEdit(this,
              m_oPop.GetTypeCodeFromName(oParentInfo.toString()),
              m_oPop.GetSpeciesCodeFromName(oInfo.toString().replace(' ', '_')));
          oEditor.pack();
          oEditor.setLocationRelativeTo(null);
          oEditor.setVisible(true);
        }
      }
      else if (oInfo.GetNodeType() == TreeNodeInfo.BEHAVIOR) {

        DisplayBehaviorEdit oEditor = new DisplayBehaviorEdit(this,
            m_oManager.GetHelpBroker());
        oEditor.pack();
        oEditor.setLocationRelativeTo(null);
        oEditor.setVisible(true);

      }
      else if (oInfo.GetNodeType() == TreeNodeInfo.TYPE) {

        //If the parent node was species, show the combo editor
        DefaultMutableTreeNode jParent = (DefaultMutableTreeNode) jNode.
            getParent();
        TreeNodeInfo oParentInfo = (TreeNodeInfo) jParent.getUserObject();
        if (oParentInfo.GetNodeType() == TreeNodeInfo.SPECIES) {

          DisplayComboEdit oEditor = new DisplayComboEdit(this,
              m_oPop.GetTypeCodeFromName(oInfo.toString()),
              m_oPop.GetSpeciesCodeFromName(oParentInfo.toString().replace(' ',
              '_')));
          oEditor.pack();
          oEditor.setLocationRelativeTo(null);
          oEditor.setVisible(true);
        }
      }
      else if (oInfo.GetNodeType() == TreeNodeInfo.GRID) {

        //Show the grid setup window
        GridSetup oEditor = new GridSetup(this, m_oManager);
        oEditor.pack();
        oEditor.setLocationRelativeTo(null);
        oEditor.setVisible(true);

      }
    }
    catch (ModelException oErr) {
      ErrorGUI oHandler = new ErrorGUI(this);
      oHandler.WriteErrorMessage(oErr);
    }
  }

  /**
   * A common function for those dialogs which contain a list of behavior
   * groups, a choice upon which populates individual behaviors into another
   * list.
   * <p>
   * If there is no choice, nothing happens.  If the behavior group is
   * "Planting" or "Output", instead of behavior choices the
   * string "--Use this behavior's own dialog" will be displayed.
   * @param jBehaviorGroupBox The combo box displaying the behavior
   * groupings.
   * @param jIndividualBehaviorListModel The list model for the individual
   * behavior choices list to be updated.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected void UpdateBehaviorChoices(JComboBox jBehaviorGroupBox,
                                       DefaultListModel
                                       jIndividualBehaviorListModel) {
    //The behavior group list was selected
    jIndividualBehaviorListModel.clear();
    int i,
        iIndex = jBehaviorGroupBox.getSelectedIndex() - 1;
    if (iIndex < 0) {
      return;
    }
    if (mp_oBehaviors[iIndex] == null || mp_oBehaviors[iIndex].length == 0) {
      jIndividualBehaviorListModel.addElement("--No behaviors");
      return;
    }
    if (mp_sBehaviorGroupNames[iIndex].equals("Planting") ||
        mp_sBehaviorGroupNames[iIndex].equals("Output")) {
      jIndividualBehaviorListModel.addElement(
          "--Use this behavior's own dialog");
      return;
    }
    for (i = 0; i < mp_oBehaviors[iIndex].length; i++) {
      jIndividualBehaviorListModel.addElement(mp_oBehaviors[iIndex][i].
                                              GetDescriptor());
    }
  }

  /**
   * Common function for those dialogs which add a behavior to a list.  This
   * extracts chosen behaviors' names, returning null if no behavior should be
   * added.  THIS DOES NOT ACTUALLY ADD THE NAME TO THE TARGET LIST.
   *
   * The behavior name is left null if there is no index selected, or if no
   * selection is eligible.  Selections are ineligible if the behavior name
   * starts with "--", or if the behavior is already on the target list.
   *
   * @param jBehaviorList The list from which the new behavior name is coming.
   * @param jBehaviorListModel The list model from which the new behavior
   * name is coming.
   * @param jListToAddTo The list that is the intended target for the new
   * behavior.
   * @return List of behavior names to add.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected String[] GetBehaviorAddName(JList jBehaviorList, DefaultListModel
                                        jBehaviorListModel, DefaultListModel
                                        jListToAddTo) {
    Vector p_oBehaviors = new Vector(0);
    int[] p_iSelectedValues = jBehaviorList.getSelectedIndices();
    int i, j;
    boolean bDuplicate;

    if (p_iSelectedValues == null || p_iSelectedValues.length == 0) {
      return null; //Nothing is selected
    }

    for (i = 0; i < p_iSelectedValues.length; i++) {
      //Get the behavior string to be added
      String sNewBehavior = (String) jBehaviorListModel.getElementAt(
          p_iSelectedValues[i]);

      //If the string starts with "--", skip it because it's a message
      //not meant for adding
      if (sNewBehavior.startsWith("--") == false) {

        //Make sure it doesn't already exist in the target list
        bDuplicate = false;
        for (j = 0; j < jListToAddTo.size(); j++) {
          if (sNewBehavior.equals(jListToAddTo.elementAt(j).toString())) {
            bDuplicate = true;
          }
        }
        if (bDuplicate == false) {
          p_oBehaviors.add(sNewBehavior);
        }
      }
    }
    if (p_oBehaviors.size() == 0) {
      return null;
    }
    String[] p_oReturn = new String[p_oBehaviors.size()];
    for (i = 0; i < p_oReturn.length; i++) {
      p_oReturn[i] = (String) p_oBehaviors.elementAt(i);
    }
    return p_oReturn;
  }

  /**
   * Objects set in tree nodes to give information about what to do when the
   * user clicks a node.
   * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
   * <p>Company: Institute of Ecosystem Studies</p>
   * @author Lora E. Murphy
   * @version 1.0
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  class TreeNodeInfo {
    /**Display name for the node*/
    String m_sName;

    /**Node type - allowed values are BEHAVIOR, SPECIES, TYPE, or IGNORE*/
    int m_iNodeType;

    public static final int BEHAVIOR = 0;
    public static final int SPECIES = 1;
    public static final int TYPE = 2;
    public static final int GRID = 3;
    public static final int IGNORE = 4;

    /**
     * Constructor.
     * @param sName Display name of the node
     * @param iType Node type - allowed values are BEHAVIOR, SPECIES, TYPE,
     * or IGNORE
     *
     * <br>Edit history:
     * <br>------------------
     * <br>April 28, 2004: Submitted in beta version (LEM)
     */
    TreeNodeInfo(String sName, int iType) {
      m_sName = sName;
      m_iNodeType = iType;
    }

    /**
     * Overridden method from java.lang.Object
     * @return Display name of node
     *
     * <br>Edit history:
     * <br>------------------
     * <br>April 28, 2004: Submitted in beta version (LEM)
     */
    public String toString() {
      return m_sName;
    }

    /**
     * Gets the node type.
     * @return Node type.
     *
     * <br>Edit history:
     * <br>------------------
     * <br>April 28, 2004: Submitted in beta version (LEM)
     */
    public int GetNodeType() {
      return m_iNodeType;
    }
  }

  /**
   * Displays the edit window for editing behavior flow for a tree species/
   * type combo.
   * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
   * <p>Company: Institute of Ecosystem Studies</p>
   * @author Lora E. Murphy
   * @version 1.0
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  class DisplayComboEdit
      extends JDialog
      implements ActionListener {

    private int m_iType, m_iSpecies;

    /**List if the behavior groupings in a combo box (Light, Growth, etc)*/
    private JComboBox m_jBehaviorGroups;

    private DefaultListModel
        /**List model of behaviors for a behavior grouping*/
        m_jBehaviorListModel,
        /**List model of behaviors applied to the species/type combo*/
        m_jComboBehaviorListModel;

    private JList
        /**List which actually displays the list of behaviors applied to the
         * species/type combo*/
        m_jComboBehaviorList,
        /**List which displays the list of behaviors for a behavior grouping*/
        m_jBehaviorList;

    private String m_sHelpID = "windows.tree_behavior_window";

    /**
     * Constructor.  Creates the dialog which allows the user to modify
     * behavior choices and behavior order for a species/type combo.
     * @param jParent Dialog window owning this dialog.
     * @param iType Tree type (life history stage).
     * @param iSpecies Tree species.
     * @throws ModelException passed through from called methods.
     *
     * <br>Edit history:
     * <br>------------------
     * <br>April 28, 2004: Submitted in beta version (LEM)
     */
    public DisplayComboEdit(JDialog jParent, int iType, int iSpecies) throws
        ModelException {
      super(jParent, true);

      m_iType = iType;
      m_iSpecies = iSpecies;

      //Help ID
      m_oManager.GetHelpBroker().enableHelpKey(this.getRootPane(), m_sHelpID, null);

      //Create a combo box with all behavior groups
      String[] p_sBehaviorGroupChoices = new String[mp_sBehaviorGroupNames.
          length + 1];
      p_sBehaviorGroupChoices[0] = "---Please select a behavior group";
      for (int i = 0; i < mp_sBehaviorGroupNames.length; i++) {
        p_sBehaviorGroupChoices[i + 1] = mp_sBehaviorGroupNames[i];
      }
      m_jBehaviorGroups = new JComboBox(p_sBehaviorGroupChoices);
      m_jBehaviorGroups.setFont(new ModelFont());
      m_jBehaviorGroups.addActionListener(this);

      //Create a list with all behaviors
      m_jBehaviorListModel = new DefaultListModel();
      m_jBehaviorList = new JList(m_jBehaviorListModel);
      m_jBehaviorList.setFont(new ModelFont());
      m_jBehaviorList.addMouseListener(new AddDoubleClicker(this));
      m_jBehaviorList.setCellRenderer(new ListRenderer(250));
      JScrollPane jScroller = new JScrollPane(m_jBehaviorList);
      jScroller.getViewport().setPreferredSize(new Dimension(250,
          (int) jScroller.getPreferredSize().
          getHeight()));

      //Package these together in a panel
      String sSpeciesName = m_oPop.GetSpeciesNameFromCode(m_iSpecies).replace(
          '_', ' '),
          sTypeName = m_oPop.GetTypeNameFromCode(m_iType);
      JPanel jLeftPanel = new JPanel();
      jLeftPanel.setLayout(new BorderLayout());
      jLeftPanel.add(m_jBehaviorGroups, BorderLayout.PAGE_START);
      jLeftPanel.add(jScroller, BorderLayout.CENTER);

      //Final choices button panel - three "apply to" choices plus cancel
      JButton jOKComboButton = new JButton("Apply to " + sSpeciesName
                                           + " " + sTypeName + "s");
      jOKComboButton.setFont(new ModelFont());
      jOKComboButton.setActionCommand("OK_Combo");
      jOKComboButton.addActionListener(this);

      JButton jOKTypeButton = new JButton("Apply to all " + sTypeName + "s");
      jOKTypeButton.setFont(new ModelFont());
      jOKTypeButton.setActionCommand("OK_Type");
      jOKTypeButton.addActionListener(this);

      JButton jCancelButton = new JButton("Cancel");
      jCancelButton.setFont(new ModelFont());
      jCancelButton.setActionCommand("Cancel");
      jCancelButton.addActionListener(this);

      JButton jHelpButton = new JButton("Help");
      jHelpButton.setFont(new ModelFont());
      jHelpButton.addActionListener(new CSH.DisplayHelpFromSource(m_oManager.
          GetHelpBroker()));
      CSH.setHelpIDString(jHelpButton, m_sHelpID);

      JPanel jButtonPanel = new JPanel();
      FlowLayout jFlowLayout = new FlowLayout();
      jFlowLayout.setAlignment(FlowLayout.RIGHT);
      jButtonPanel.setLayout(jFlowLayout);
      jButtonPanel.add(jOKComboButton);
      jButtonPanel.add(jOKTypeButton);
      jButtonPanel.add(jCancelButton);
      jButtonPanel.add(jHelpButton);
      jButtonPanel.setBorder(BorderFactory.createMatteBorder(1, 0, 0, 0,
          Color.BLACK));

      //Label displaying the type and species
      JLabel jLabel = new JLabel(m_oPop.GetSpeciesNameFromCode(m_iSpecies).
                                 replace('_', ' ')
                                 + " " + m_oPop.GetTypeNameFromCode(m_iType) +
                                 "s");
      jLabel.setFont(new Font("Dialog", Font.BOLD, 16));

      //List showing this combo's behaviors
      m_jComboBehaviorListModel = new DefaultListModel();
      m_jComboBehaviorList = new JList(m_jComboBehaviorListModel);
      m_jComboBehaviorList.setCellRenderer(new ListRenderer(250));
      JScrollPane jBehaviorScroller = new JScrollPane(m_jComboBehaviorList);
      jBehaviorScroller.getViewport().setPreferredSize(new Dimension(250,
          (int) jBehaviorScroller.getViewport().getPreferredSize().getHeight()));
      m_jComboBehaviorList.setFont(new ModelFont());
      int i;
      Vector oComboBehaviors = GetBehaviorList(m_iSpecies, m_iType);
      for (i = 0; i < oComboBehaviors.size(); i++) {
        String sBehaviorName = (String) oComboBehaviors.elementAt(i);
        m_jComboBehaviorListModel.addElement(new BehaviorIndexer(
            sBehaviorName, GetBehaviorIndex(sBehaviorName)));
      }

      //Label for the behavior list
      JLabel jBehaviorLabel = new JLabel("Assigned behaviors:");
      jBehaviorLabel.setFont(new ModelFont());
      //Layout behavior into a panel
      JPanel jBehaviorPanel = new JPanel();
      jBehaviorPanel.setLayout(new BorderLayout());
      jBehaviorPanel.add(jBehaviorScroller, BorderLayout.CENTER);
      jBehaviorPanel.add(jBehaviorLabel, BorderLayout.PAGE_START);

      JButton jRemoveButton = new JButton("Remove");
      jRemoveButton.setFont(new ModelFont());
      jRemoveButton.setToolTipText("Remove this behavior from this tree type");
      jRemoveButton.setActionCommand("Remove");
      jRemoveButton.addActionListener(this);

      //Put the buttons in the panel
      JPanel jChangeButtonPanel = new JPanel();
      jChangeButtonPanel.setLayout(new GridLayout(0, 1));
      jChangeButtonPanel.add(jRemoveButton);

      //Button and panel to add behaviors to combo
      JButton jAddButton = new JButton(new ModelIcon(15, 15,
          ModelIcon.RIGHT_TRIANGLE));
      jAddButton.setFont(new ModelFont());
      jAddButton.setToolTipText("Add selected behavior");
      jAddButton.setActionCommand("Add");
      jAddButton.addActionListener(this);

      JPanel jAddButtonPanel = new JPanel();
      jAddButtonPanel.setLayout(new BorderLayout());
      jAddButtonPanel.add(jAddButton, BorderLayout.PAGE_START);
      jAddButtonPanel.add(new JPanel(), BorderLayout.CENTER);

      //Put it all together
      JPanel jPackager = new JPanel();
      jPackager.setLayout(new FlowLayout());
      jPackager.add(jLeftPanel);
      jPackager.add(jAddButtonPanel);
      jPackager.add(jBehaviorPanel);
      jPackager.add(jChangeButtonPanel);

      getContentPane().setLayout(new BorderLayout());
      getContentPane().add(jButtonPanel, BorderLayout.PAGE_END);
      getContentPane().add(jPackager, BorderLayout.CENTER);
      getContentPane().add(jLabel, BorderLayout.PAGE_START);
    }

    /**
     * Gets the index for a behavior (the absolute order in which behaviors
     * go).
     * @param sBehaviorName Behavior name to get index for.
     * @return Behavior index.
     *
     * <br>Edit history:
     * <br>------------------
     * <br>April 28, 2004: Submitted in beta version (LEM)
     */
    protected int GetBehaviorIndex(String sBehaviorName) {
      int i, j,
          iIndex = 0;
      for (i = 0; i < mp_oBehaviors.length; i++) {
        if (mp_oBehaviors[i] != null) {
          for (j = 0; j < mp_oBehaviors[i].length; j++) {
            iIndex++;
            if (mp_oBehaviors[i][j].GetDescriptor().equals(sBehaviorName)) {
              return iIndex;
            }
          }
        }
      }
      return -1;
    }

    /**
     * Performs actions for clicked buttons.
     * @param e ActionEvent object.
     *
     * <br>Edit history:
     * <br>------------------
     * <br>April 28, 2004: Submitted in beta version (LEM)
     */
    public void actionPerformed(ActionEvent e) {
      if (e.getActionCommand().equals("OK_Combo")) {
        try {
          //Set the behaviors for this combo
          int iNumBehaviors = 0, i;
          for (i = 0; i < m_jComboBehaviorListModel.getSize(); i++) {
            if (m_jComboBehaviorListModel.elementAt(i) instanceof
                BehaviorIndexer) {
              iNumBehaviors++;
            }
          }
          String[] p_sBehaviors = new String[iNumBehaviors];
          iNumBehaviors = 0;
          for (i = 0; i < m_jComboBehaviorListModel.getSize(); i++) {
            if (m_jComboBehaviorListModel.elementAt(i) instanceof
                BehaviorIndexer) {
              p_sBehaviors[iNumBehaviors] = m_jComboBehaviorListModel.elementAt(
                  i).toString();
              iNumBehaviors++;
            }
          }

          SetBehaviorList(m_iSpecies, m_iType, p_sBehaviors);

          this.setVisible(false);
          this.dispose();
        }
        catch (ModelException oErr) {
          ErrorGUI oHandler = new ErrorGUI(this);
          oHandler.WriteErrorMessage(oErr);
        }
      }
      else if (e.getActionCommand().equals("OK_Species")) {
        try {
          //Set the behaviors for all types for this species
          TreePopulation oPop = m_oManager.GetTreePopulation();
          int iNumBehaviors = 0, i,
              iNumTypes = oPop.GetNumberOfTypes();
          for (i = 0; i < m_jComboBehaviorListModel.getSize(); i++) {
            if (m_jComboBehaviorListModel.elementAt(i) instanceof
                BehaviorIndexer) {
              iNumBehaviors++;
            }
          }
          String[] p_sBehaviors = new String[iNumBehaviors];
          iNumBehaviors = 0;
          for (i = 0; i < m_jComboBehaviorListModel.getSize(); i++) {
            if (m_jComboBehaviorListModel.elementAt(i) instanceof
                BehaviorIndexer) {
              p_sBehaviors[iNumBehaviors] = m_jComboBehaviorListModel.elementAt(
                  i).toString();
              iNumBehaviors++;
            }
          }

          for (i = 0; i < iNumTypes; i++) {
            SetBehaviorList(m_iSpecies, i, p_sBehaviors);
          }

          this.setVisible(false);
          this.dispose();
        }
        catch (ModelException oErr) {
          ErrorGUI oHandler = new ErrorGUI(this);
          oHandler.WriteErrorMessage(oErr);
        }
      }
      else if (e.getActionCommand().equals("OK_Type")) {
        try {
          //Set the behaviors for all species for this type
          TreePopulation oPop = m_oManager.GetTreePopulation();
          int iNumSpecies = oPop.GetNumberOfSpecies(), i,
              iNumBehaviors = 0;

          for (i = 0; i < m_jComboBehaviorListModel.getSize(); i++) {
            if (m_jComboBehaviorListModel.elementAt(i) instanceof
                BehaviorIndexer) {
              iNumBehaviors++;
            }
          }
          String[] p_sBehaviors = new String[iNumBehaviors];
          iNumBehaviors = 0;
          for (i = 0; i < m_jComboBehaviorListModel.getSize(); i++) {
            if (m_jComboBehaviorListModel.elementAt(i) instanceof
                BehaviorIndexer) {
              p_sBehaviors[iNumBehaviors] = m_jComboBehaviorListModel.elementAt(
                  i).toString();
              iNumBehaviors++;
            }
          }

          for (i = 0; i < iNumSpecies; i++) {
            SetBehaviorList(i, m_iType, p_sBehaviors);
          }

          this.setVisible(false);
          this.dispose();
        }
        catch (ModelException oErr) {
          ErrorGUI oHandler = new ErrorGUI(this);
          oHandler.WriteErrorMessage(oErr);
        }
      }
      else if (e.getActionCommand().equals("Remove")) {

        //Remove the selected behavior from the list
        Object[] p_oSelected = m_jComboBehaviorList.getSelectedValues();

        if (p_oSelected == null || p_oSelected.length == 0) {
          JOptionPane.showMessageDialog(this, "Please select a behavior.");
          return;
        }

        for (int i = 0; i < p_oSelected.length; i++) {
          m_jComboBehaviorListModel.removeElement(p_oSelected[i]);
        }
      }
      else if (e.getActionCommand().equals("Add")) {

        int i, j;

        //Get the behavior to be added
        String[] p_sNewBehaviors = GetBehaviorAddName(m_jBehaviorList,
            m_jBehaviorListModel,
            m_jComboBehaviorListModel);

        if (p_sNewBehaviors == null || p_sNewBehaviors.length == 0) {
          JOptionPane.showMessageDialog(this, "Please select a behavior.");
          return;
        }

        BehaviorIndexer oNewBehavior;

        //OK, now add it in the proper order
        BehaviorIndexer oExistingBehavior;
        if (m_jComboBehaviorListModel.size() == 0) {
          //First behavior
          for (i = 0; i < p_sNewBehaviors.length; i++) {
            oNewBehavior = new BehaviorIndexer(p_sNewBehaviors[i], GetBehaviorIndex(p_sNewBehaviors[i]));
            m_jComboBehaviorListModel.addElement(oNewBehavior);
          }
          return;
        }
        for (i = 0; i < m_jComboBehaviorListModel.size(); i++) {
          oNewBehavior = new BehaviorIndexer(p_sNewBehaviors[0], GetBehaviorIndex(p_sNewBehaviors[0]));
          oExistingBehavior = (BehaviorIndexer) m_jComboBehaviorListModel.
              elementAt(i);
          if (oExistingBehavior.m_iIndex > oNewBehavior.m_iIndex) {

            //The new behavior belongs just before
            for (j = 0; j < p_sNewBehaviors.length; j++) {
              oNewBehavior = new BehaviorIndexer(p_sNewBehaviors[j], GetBehaviorIndex(p_sNewBehaviors[j]));
              m_jComboBehaviorListModel.insertElementAt(oNewBehavior, i);
            }
            return;
          }
        }

        //If we're still here, put this last
        for (i = 0; i < p_sNewBehaviors.length; i++) {
          oNewBehavior = new BehaviorIndexer(p_sNewBehaviors[i], GetBehaviorIndex(p_sNewBehaviors[i]));
          m_jComboBehaviorListModel.addElement(oNewBehavior);
        }
      }
      else if (e.getActionCommand().equals("Cancel")) {
        this.setVisible(false);
        this.dispose();
      }
      else if (e.getSource().equals(m_jBehaviorGroups)) {
        UpdateBehaviorChoices(m_jBehaviorGroups, m_jBehaviorListModel);
      }
    }

    /**This class implements the double-click so that double-clicking on
     * behavior names adds the behavior to the list*/
    class AddDoubleClicker
        implements MouseListener {
      ActionListener m_jDialog;
      public AddDoubleClicker(ActionListener jDialog) {
        m_jDialog = jDialog;
      }

      public void mouseClicked(MouseEvent e) {
        if (e.getClickCount() == 2) {
          m_jDialog.actionPerformed(new ActionEvent(this, 0, "Add"));
        }
      }

      public void mousePressed(MouseEvent e) {
        ;
      }

      public void mouseEntered(MouseEvent e) {
        ;
      }

      public void mouseExited(MouseEvent e) {
        ;
      }

      public void mouseReleased(MouseEvent e) {
        ;
      }
    }
  }

  /**
   * Displays the edit window for editing behavior order.
   * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
   * <p>Company: Institute of Ecosystem Studies</p>
   * @author Lora E. Murphy
   * @version 1.0
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  class DisplayBehaviorEdit
      extends JDialog
      implements ActionListener {

    /**List if the behavior groupings in a combo box (Light, Growth, etc)*/
    private JComboBox m_jBehaviorGroups;

    private DefaultListModel
        /**List model of behaviors for a behavior grouping*/
        m_jBehaviorListModel,
        /**List model of behaviors currently enabled*/
        m_jEnabledBehaviorListModel;

    private JList
        /**List which actually displays the list of behaviors currently
         * enabled*/
        m_jEnabledBehaviorList,
        /**List which displays the list of behaviors for a behavior grouping*/
        m_jBehaviorList;

    private String m_sHelpID = "windows.run_behaviors_window";

    /**
     * Constructor.  Creates the GUI.
     * @param jParent Parent dialog in which to display this one.
     * @param oHelpBroker System Help broker.
     *
     * <br>Edit history:
     * <br>------------------
     * <br>April 28, 2004: Submitted in beta version (LEM)
     */
    public DisplayBehaviorEdit(JDialog jParent, HelpBroker oHelpBroker) {
      super(jParent, "Current Run Behaviors", true);

      //Help ID
      oHelpBroker.enableHelpKey(this.getRootPane(), m_sHelpID, null);

      //Create a combo box with all behavior groups
      String[] p_sBehaviorGroupChoices = new String[mp_sBehaviorGroupNames.
          length + 1];
      p_sBehaviorGroupChoices[0] = "---Please select a behavior group";
      for (int i = 0; i < mp_sBehaviorGroupNames.length; i++) {
        p_sBehaviorGroupChoices[i + 1] = mp_sBehaviorGroupNames[i];
      }
      m_jBehaviorGroups = new JComboBox(p_sBehaviorGroupChoices);
      m_jBehaviorGroups.setFont(new ModelFont());
      m_jBehaviorGroups.addActionListener(this);

      //Create a list with all behaviors
      m_jBehaviorListModel = new DefaultListModel();
      m_jBehaviorList = new JList(m_jBehaviorListModel);
      m_jBehaviorList.setCellRenderer(new ListRenderer(250));
      m_jBehaviorList.addMouseListener(new AddDoubleClicker(this));
      m_jBehaviorList.setFont(new ModelFont());
      JScrollPane jScroller = new JScrollPane(m_jBehaviorList);
      jScroller.getViewport().setPreferredSize(new Dimension(250,
          (int) jScroller.getViewport().getPreferredSize().
          getHeight()));

      //Package these together in a panel
      JPanel jLeftPanel = new JPanel();
      jLeftPanel.setLayout(new BorderLayout());
      jLeftPanel.add(m_jBehaviorGroups, BorderLayout.PAGE_START);
      jLeftPanel.add(jScroller, BorderLayout.CENTER);

      //List showing currently enabled behaviors
      m_jEnabledBehaviorListModel = new DefaultListModel();
      m_jEnabledBehaviorList = new JList(m_jEnabledBehaviorListModel);
      //Set double-click to launch the behavior assignment window
      m_jEnabledBehaviorList.addMouseListener(new ModifyDoubleClicker(this));
      JScrollPane jBehaviorScroller = new JScrollPane(m_jEnabledBehaviorList);
      jBehaviorScroller.getViewport().setPreferredSize(new Dimension(250,
          (int) jBehaviorScroller.getViewport().getPreferredSize().
          getHeight()));
      m_jEnabledBehaviorList.setFont(new ModelFont());
      m_jEnabledBehaviorList.setCellRenderer(new ListRenderer(250));

      int i;
      Vector oEnabledBehaviors = GetEnabledBehaviorList();
      for (i = 0; i < oEnabledBehaviors.size(); i++) {
        String sBehaviorName = (String) oEnabledBehaviors.elementAt(i);
        m_jEnabledBehaviorListModel.addElement(new BehaviorIndexer(
            sBehaviorName, GetBehaviorGroupIndex(sBehaviorName)));
      }

      //Insert breaks between group indexes
      for (i = 0; i < m_jEnabledBehaviorListModel.size(); i++) {
        BehaviorIndexer oThis = (BehaviorIndexer) m_jEnabledBehaviorListModel.
            elementAt(i);
        if (i > 0) {
          if (m_jEnabledBehaviorListModel.elementAt(i - 1) instanceof
              BehaviorIndexer) {
            BehaviorIndexer oPrevious = (BehaviorIndexer)
                m_jEnabledBehaviorListModel.elementAt(i - 1);
            if (oPrevious.m_iIndex < oThis.m_iIndex) {
              m_jEnabledBehaviorListModel.insertElementAt("-----------------",
                  i);
              i++;
            }
          }
        }
      }

      //Label for the behavior list
      JLabel jBehaviorLabel = new JLabel("Current behavior order:");
      jBehaviorLabel.setFont(new ModelFont());
      //Layout behavior into a panel
      JPanel jBehaviorPanel = new JPanel();
      jBehaviorPanel.setLayout(new BorderLayout());
      jBehaviorPanel.add(jBehaviorScroller, BorderLayout.CENTER);
      jBehaviorPanel.add(jBehaviorLabel, BorderLayout.PAGE_START);

      //Buttons and panel to change order
      JButton jDownButton = new JButton("Down");
      jDownButton.setFont(new ModelFont());
      jDownButton.setToolTipText("Move selected behavior down in order");
      jDownButton.setActionCommand("Down");
      jDownButton.addActionListener(this);

      JButton jUpButton = new JButton("Up");
      jUpButton.setFont(new ModelFont());
      jUpButton.setToolTipText("Move selected behavior up in order");
      jUpButton.setActionCommand("Up");
      jUpButton.addActionListener(this);

      JButton jRemoveButton = new JButton("Remove");
      jRemoveButton.setFont(new ModelFont());
      jRemoveButton.setToolTipText("Remove behavior from run");
      jRemoveButton.setActionCommand("Remove");
      jRemoveButton.addActionListener(this);

      JButton jModifyButton = new JButton("Modify assigned data");
      jModifyButton.setFont(new ModelFont());
      jModifyButton.setToolTipText(
          "Set the trees to which this behavior applies");
      jModifyButton.setActionCommand("Modify");
      jModifyButton.addActionListener(this);

      //Put the buttons in the panel
      JPanel jChangeButtonPanel = new JPanel();
      jChangeButtonPanel.setLayout(new GridLayout(0, 1));
      jChangeButtonPanel.add(jUpButton);
      jChangeButtonPanel.add(jDownButton);
      jChangeButtonPanel.add(jRemoveButton);
      jChangeButtonPanel.add(jModifyButton);

      //Button and panel to add behaviors to the list
      JButton jAddButton = new JButton(new ModelIcon(15, 15,
          ModelIcon.RIGHT_TRIANGLE));
      jAddButton.setToolTipText("Add selected behavior");
      jAddButton.setFont(new ModelFont());
      jAddButton.setActionCommand("Add");
      jAddButton.addActionListener(this);

      JPanel jAddButtonPanel = new JPanel();
      jAddButtonPanel.setLayout(new BorderLayout());
      jAddButtonPanel.add(jAddButton, BorderLayout.PAGE_START);
      jAddButtonPanel.add(new JPanel(), BorderLayout.CENTER);

      //Put it all together
      JPanel jPackager = new JPanel();
      jPackager.setLayout(new FlowLayout());
      jPackager.add(jLeftPanel);
      jPackager.add(jAddButtonPanel);
      jPackager.add(jBehaviorPanel);
      jPackager.add(jChangeButtonPanel);

      getContentPane().setLayout(new BorderLayout());
      getContentPane().add(new OKCancelButtonPanel(this, oHelpBroker, m_sHelpID),
                           BorderLayout.PAGE_END);
      getContentPane().add(jPackager, BorderLayout.CENTER);
    }

    /**
     * Gets a list of enabled behavior names.
     * @return List of enabled behavior names.
     *
     * <br>Edit history:
     * <br>------------------
     * <br>April 28, 2004: Submitted in beta version (LEM)
     */
    protected Vector GetEnabledBehaviorList() {
      Vector oBehaviors = new Vector(0);
      int i, j;

      for (i = 0; i < mp_oBehaviors.length; i++) {
        for (j = 0; j < mp_oBehaviors[i].length; j++) {
          if (mp_oBehaviors[i][j].m_bIsEnabled) {
            oBehaviors.add(mp_oBehaviors[i][j].GetDescriptor());
          }
        }
      }

      return oBehaviors;
    }

    /**
     * Gets the group index for a behavior (the order in which parent
     * behavior groups go).  This keeps behaviors in group order so they can
     * only move within their group.
     * @param sBehaviorName Behavior name to get index for.
     * @return Behavior index.
     *
     * <br>Edit history:
     * <br>------------------
     * <br>April 28, 2004: Submitted in beta version (LEM)
     */
    protected int GetBehaviorGroupIndex(String sBehaviorName) {
      int i, j;
      for (i = 0; i < mp_oBehaviors.length; i++) {
        if (mp_oBehaviors[i] != null) {
          for (j = 0; j < mp_oBehaviors[i].length; j++) {
            if (mp_oBehaviors[i][j].GetDescriptor().equals(sBehaviorName)) {
              return i;
            }
          }
        }
      }
      return -1;
    }

    /**
     * Performs actions for clicked buttons.
     * @param e ActionEvent object.
     *
     * <br>Edit history:
     * <br>------------------
     * <br>April 28, 2004: Submitted in beta version (LEM)
     */
    public void actionPerformed(ActionEvent e) {
      if (e.getActionCommand().equals("Up")) {

        //Move the selected behavior up in the list - but it can't
        //go to a smaller group index
        int iSelectedIndex = m_jEnabledBehaviorList.getSelectedIndex(), //selection
            iNewIndex = iSelectedIndex - 1; //proposed new position of item

        //If nothing selected or the first item is selected, warn and exit
        if ( -1 == iSelectedIndex || 0 == iSelectedIndex) {
          JOptionPane.showMessageDialog(this, "Please select a behavior.");
          return;
        }

        //If the position immediately above is a String (and thus a dividing
        //line between group indexes), do nothing
        if (m_jEnabledBehaviorListModel.getElementAt(iNewIndex) instanceof
            String) {
          return;
        }

        //Swap the object from old position to new
        Object oItem = m_jEnabledBehaviorListModel.getElementAt(iSelectedIndex);
        m_jEnabledBehaviorListModel.insertElementAt(oItem, iNewIndex);
        m_jEnabledBehaviorListModel.removeElementAt(iSelectedIndex + 1);

        //Keep the same item selected
        m_jEnabledBehaviorList.setSelectedIndex(iNewIndex);

      }
      else if (e.getActionCommand().equals("Remove")) {

        //Remove the selected behavior(s) from the list
        Object[] p_oAllSelected = m_jEnabledBehaviorList.getSelectedValues();
        int iSelectedIndex, i, j;
        if (p_oAllSelected == null || p_oAllSelected.length == 0) {
          JOptionPane.showMessageDialog(this, "Please select a behavior.");
          return;
        }

        for (i = 0; i < p_oAllSelected.length; i++) {
          //Get the index of this item
          iSelectedIndex = -1;
          for (j = 0; j < m_jEnabledBehaviorListModel.size(); j++) {
            if (p_oAllSelected[i].equals(m_jEnabledBehaviorListModel.elementAt(
                j))) {
              iSelectedIndex = j;
            }
          }
          if ( -1 != iSelectedIndex &&
              ! (m_jEnabledBehaviorListModel.elementAt(iSelectedIndex) instanceof
                 String)) {

            //If this was the only item in its group index, remove a divider if
            //appropriate

            if (iSelectedIndex == 0) {
              //First item in list - is the second item a divider?
              if (m_jEnabledBehaviorListModel.size() > 1 &&
                  m_jEnabledBehaviorListModel.elementAt(iSelectedIndex + 1) instanceof
                  String) {
                m_jEnabledBehaviorListModel.removeElementAt(iSelectedIndex + 1);
              }
              m_jEnabledBehaviorListModel.removeElementAt(iSelectedIndex);

            }
            else if (iSelectedIndex == m_jEnabledBehaviorListModel.size() - 1) {
              //Last item on the list
              if (m_jEnabledBehaviorListModel.elementAt(iSelectedIndex - 1) instanceof
                  String) {
                //Remove last two
                m_jEnabledBehaviorListModel.removeElementAt(
                    m_jEnabledBehaviorListModel.
                    size() - 1);
                m_jEnabledBehaviorListModel.removeElementAt(
                    m_jEnabledBehaviorListModel.
                    size() - 1);
              }
              else {
                m_jEnabledBehaviorListModel.removeElementAt(iSelectedIndex);
              }
            }
            else {
              //Somewhere in the middle - if bracketed by strings remove one
              if (m_jEnabledBehaviorListModel.elementAt(iSelectedIndex - 1) instanceof
                  String &&
                  m_jEnabledBehaviorListModel.elementAt(iSelectedIndex + 1) instanceof
                  String) {
                m_jEnabledBehaviorListModel.removeElementAt(iSelectedIndex + 1);
              }
              m_jEnabledBehaviorListModel.removeElementAt(iSelectedIndex);
            }
          }
        }
      }
      else if (e.getActionCommand().equals("Down")) {

        //Move the selected behavior down in the list - but it can't
        //go to a larger group index
        int iSelectedIndex = m_jEnabledBehaviorList.getSelectedIndex(), //selection
            iNewIndex = iSelectedIndex + 2; //new position of item

        //If nothing selected or the first item is selected, exit
        if ( -1 == iSelectedIndex ||
            (m_jEnabledBehaviorListModel.getSize() - 1) == iSelectedIndex) {
          JOptionPane.showMessageDialog(this, "Please select a behavior.");
          return;
        }

        //If the position immediately below is a String (and thus a dividing
        //line between group indexes), do nothing
        if (m_jEnabledBehaviorListModel.getElementAt(iSelectedIndex + 1) instanceof
            String) {
          return;
        }

        //Swap the object from old position to new
        Object oItem = m_jEnabledBehaviorListModel.getElementAt(iSelectedIndex);
        m_jEnabledBehaviorListModel.insertElementAt(oItem, iNewIndex);
        m_jEnabledBehaviorListModel.removeElementAt(iSelectedIndex);

        //Keep the same item selected
        m_jEnabledBehaviorList.setSelectedIndex(iNewIndex - 1);
      }
      else if (e.getActionCommand().equals("Add")) {

        //Get the behavior to be added
        String[] sNewBehaviors = GetBehaviorAddName(m_jBehaviorList,
            m_jBehaviorListModel,
            m_jEnabledBehaviorListModel);

        if (sNewBehaviors == null) {
          JOptionPane.showMessageDialog(this, "Please select a behavior.");
          return;
        }

        int iGroupIndex = GetBehaviorGroupIndex(sNewBehaviors[0]);

        BehaviorIndexer oNewBehavior;

        //Find where to put these new behaviors
        int i;
        BehaviorIndexer oExistingBehavior;
        if (m_jEnabledBehaviorListModel.size() == 0) {
          //First behavior
          for (int j = 0; j < sNewBehaviors.length; j++) {
            oNewBehavior = new BehaviorIndexer(sNewBehaviors[j], iGroupIndex);
            m_jEnabledBehaviorListModel.addElement(oNewBehavior);
          }
          //Set the selected index to the last newly added behavior
          m_jEnabledBehaviorList.setSelectedIndex(m_jEnabledBehaviorListModel.
                                                  size() - 1);
          return;
        }
        for (i = 0; i < m_jEnabledBehaviorListModel.size(); i++) {
          if (m_jEnabledBehaviorListModel.elementAt(i) instanceof
              BehaviorIndexer) {
            oExistingBehavior = (BehaviorIndexer) m_jEnabledBehaviorListModel.
                elementAt(i);
            if (oExistingBehavior.m_iIndex == iGroupIndex) {

              //We've found these behaviors' group - put them first
              for (int j = 0; j < sNewBehaviors.length; j++) {
                oNewBehavior = new BehaviorIndexer(sNewBehaviors[j],
                    iGroupIndex);
                m_jEnabledBehaviorListModel.insertElementAt(oNewBehavior, i);
              }
              //Set this as selected
              m_jEnabledBehaviorList.setSelectedIndex(i);
              return;
            }
            if (oExistingBehavior.m_iIndex > iGroupIndex) {

              //The new behavior belongs just before - put it there with
              //a divider immediately after
              m_jEnabledBehaviorListModel.insertElementAt("-----------------",
                  i);
              for (int j = 0; j < sNewBehaviors.length; j++) {
                oNewBehavior = new BehaviorIndexer(sNewBehaviors[j],
                    iGroupIndex);
                m_jEnabledBehaviorListModel.insertElementAt(oNewBehavior, i);
              }
              //Set the selected index to the newly added one
              m_jEnabledBehaviorList.setSelectedIndex(i);
              return;
            }
          }
        }

        //If we're still here, put this last with a divider just before
        m_jEnabledBehaviorListModel.addElement("-----------------");
        for (int j = 0; j < sNewBehaviors.length; j++) {
          oNewBehavior = new BehaviorIndexer(sNewBehaviors[j], iGroupIndex);
          m_jEnabledBehaviorListModel.addElement(oNewBehavior);
        }
        m_jEnabledBehaviorList.setSelectedIndex(m_jEnabledBehaviorListModel.
                                                size() - 1);
      }
      else if (e.getActionCommand().equals("Modify")) {
        try {

          //Call up a dialog to modify a chosen individual behavior's
          //tree assignments
          int iSelectedIndex = m_jEnabledBehaviorList.getSelectedIndex();

          if ( -1 == iSelectedIndex) {
            JOptionPane.showMessageDialog(this, "Please select a behavior.");
            return;
          }

          if (m_jEnabledBehaviorListModel.elementAt(iSelectedIndex) instanceof
              String) {
            JOptionPane.showMessageDialog(this, "Please select a behavior.");
            return;
          }

          String sBehaviorName =
              m_jEnabledBehaviorListModel.elementAt(iSelectedIndex).toString();

          DisplayBehaviorComboEdit oEditor = new DisplayBehaviorComboEdit(this,
              sBehaviorName, m_oManager.GetHelpBroker());

          oEditor.pack();
          oEditor.setLocationRelativeTo(null);
          oEditor.setVisible(true);
        }
        catch (ModelException oErr) {
          ErrorGUI oHandler = new ErrorGUI(this);
          oHandler.WriteErrorMessage(oErr);
        }
      }
      else if (e.getActionCommand().equals("Cancel")) {
        this.setVisible(false);
        this.dispose();
      }
      else if (e.getActionCommand().equals("OK")) {
        try {

          //Set the enabled behaviors to the ones on the list
          int i, j, k;

          //First, disable all behaviors
          for (i = 0; i < mp_oBehaviors.length; i++) {
            for (j = 0; j < mp_oBehaviors[i].length; j++) {
              mp_oBehaviors[i][j].m_bIsEnabled = false;
            }
          }

          //Now set the enabled behaviors to the ones currently visible in
          //the list
          String sBehaviorName;
          for (k = 0; k < m_jEnabledBehaviorListModel.size(); k++) {
            if (m_jEnabledBehaviorListModel.elementAt(k) instanceof
                BehaviorIndexer) {
              sBehaviorName = m_jEnabledBehaviorListModel.elementAt(k).toString();

              //Find this behavior and set it to enabled
              for (i = 0; i < mp_oBehaviors.length; i++) {
                for (j = 0; j < mp_oBehaviors[i].length; j++) {
                  if (mp_oBehaviors[i][j].GetDescriptor().equals(sBehaviorName)) {
                    mp_oBehaviors[i][j].m_bIsEnabled = true;
                  }
                }
              }
            }
          }

          //Re-do the order
          //Start by building an array of vectors - the array holds one vector
          //for each behavior group, and each vector contains the enabled
          //behavior names in order
          BehaviorIndexer oIndexedBehavior;
          Vector[] p_oNewBehaviorOrder = new Vector[mp_oBehaviors.length];
          for (i = 0; i < p_oNewBehaviorOrder.length; i++) {
            p_oNewBehaviorOrder[i] = new Vector(0);
          }
          for (i = 0; i < m_jEnabledBehaviorListModel.size(); i++) {
            if (m_jEnabledBehaviorListModel.elementAt(i) instanceof
                BehaviorIndexer) {
              oIndexedBehavior = (BehaviorIndexer) m_jEnabledBehaviorListModel.
                  elementAt(i);
              p_oNewBehaviorOrder[oIndexedBehavior.m_iIndex].add(
                  oIndexedBehavior.m_sBehaviorName);
            }
          }

          //Now go through each group and reset the behaviors accordingly
          for (i = 0; i < p_oNewBehaviorOrder.length; i++) {

            if (p_oNewBehaviorOrder[i].size() > 0) {

              //Create a copy of the existing array and null out the current
              //array
              Behavior[] p_oCopy = new Behavior[mp_oBehaviors[i].length];
              //This tells us which behaviors have been copied over
              boolean[] p_oWasCopied = new boolean[p_oCopy.length];
              int iCopyIndex = 0;
              for (j = 0; j < p_oCopy.length; j++) {
                p_oCopy[j] = mp_oBehaviors[i][j];
                mp_oBehaviors[i][j] = null;
                p_oWasCopied[j] = false;
              }

              //Go through the Vector and put its behaviors in order
              for (j = 0; j < p_oNewBehaviorOrder[i].size(); j++) {
                sBehaviorName = (String) p_oNewBehaviorOrder[i].elementAt(j);
                for (k = 0; k < p_oCopy.length; k++) {
                  if (sBehaviorName.equals(p_oCopy[k].GetDescriptor())) {
                    mp_oBehaviors[i][j] = p_oCopy[k];
                    p_oWasCopied[k] = true;
                  }
                }
              }

              //Now any that didn't get copied (because they were disabled),
              //add to the end
              iCopyIndex = p_oNewBehaviorOrder[i].size();
              for (j = 0; j < p_oWasCopied.length; j++) {
                if (p_oWasCopied[j] == false) {
                  mp_oBehaviors[i][iCopyIndex] = p_oCopy[j];
                  iCopyIndex++;
                }
              }
            }
          }

          BuildTree();

          this.setVisible(false);
          this.dispose();
        }
        catch (ModelException oErr) {
          ErrorGUI oHandler = new ErrorGUI(this);
          oHandler.WriteErrorMessage(oErr);
        }
      }
      else if (e.getSource().equals(m_jBehaviorGroups)) {
        //The behavior group list was selected
        UpdateBehaviorChoices(m_jBehaviorGroups, m_jBehaviorListModel);
      }
    }

    /**This class implements the double-click so that double-clicking on
     * behavior names launches the behavior combo editor*/
    class ModifyDoubleClicker
        implements MouseListener {
      ActionListener m_jDialog;
      public ModifyDoubleClicker(ActionListener jDialog) {
        m_jDialog = jDialog;
      }

      public void mouseClicked(MouseEvent e) {
        if (e.getClickCount() == 2) {
          m_jDialog.actionPerformed(new ActionEvent(this, 0, "Modify"));
        }
      }

      public void mousePressed(MouseEvent e) {
        ;
      }

      public void mouseEntered(MouseEvent e) {
        ;
      }

      public void mouseExited(MouseEvent e) {
        ;
      }

      public void mouseReleased(MouseEvent e) {
        ;
      }
    }

    /**This class implements the double-click so that double-clicking on
     * behavior names adds the behavior to the list*/
    class AddDoubleClicker
        implements MouseListener {
      ActionListener m_jDialog;
      public AddDoubleClicker(ActionListener jDialog) {
        m_jDialog = jDialog;
      }

      public void mouseClicked(MouseEvent e) {
        if (e.getClickCount() == 2) {
          m_jDialog.actionPerformed(new ActionEvent(this, 0, "Add"));
        }
      }

      public void mousePressed(MouseEvent e) {
        ;
      }

      public void mouseEntered(MouseEvent e) {
        ;
      }

      public void mouseExited(MouseEvent e) {
        ;
      }

      public void mouseReleased(MouseEvent e) {
        ;
      }
    }
  }

  /**Class which packages individual behavior names with an index number.
   * The index can be used to set and enforce behavior orders.
   * @author Lora E. Murphy
   *
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  private class BehaviorIndexer {

    /**
     * Constructor.
     * @param sBehaviorName Name of behavior.
     * @param iIndex Index number to package with name.
     *
     * <br>Edit history:
     * <br>------------------
     * <br>April 28, 2004: Submitted in beta version (LEM)
     */
    BehaviorIndexer(String sBehaviorName, int iIndex) {
      m_sBehaviorName = sBehaviorName;
      m_iIndex = iIndex;
    }

    public String toString() {
      return m_sBehaviorName;
    }

    String m_sBehaviorName; /**<Behavior display name*/
    int m_iIndex; /**<Index to be saved with behavior*/
  }

  /**
   * Displays an edit window for editing the tree species/type combos to
   * which a single behavior applies.
   * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
   * <p>Company: Institute of Ecosystem Studies</p>
   * @author Lora E. Murphy
   * @version 1.0
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  private class DisplayBehaviorComboEdit
      extends JDialog
      implements ActionListener {

    /**The list displaying the assigned species/type combos for this
     * behavior*/
    protected JList m_jAssignedCombosList;

    /**The list of species*/
    protected JList m_jSpecies;

    /**The list of types*/
    protected JList m_jTypes;

    /**The assigned species/type combos for this behavior*/
    protected DefaultListModel m_jAssignedCombosListModel;

    /**The name of the behavior*/
    protected String m_sBehaviorName;

    /**The help topic ID for this window*/
    private String m_sHelpID = "windows.behavior_assignments";

    /**
     * Constructor.
     * @param jParent Parent dialog in which to display this dialog.
     * @param sBehaviorName Behavior to edit.
     * @param oHelpBroker The application's help broker.
     * @throws ModelException if there is a problem constructing the window.
     *
     * <br>Edit history:
     * <br>------------------
     * <br>April 28, 2004: Submitted in beta version (LEM)
     */
    public DisplayBehaviorComboEdit(JDialog jParent, String sBehaviorName,
                                    HelpBroker oHelpBroker) throws
        ModelException {
      super(jParent, "Tree assignments for " + sBehaviorName, true);

      //Help ID
      oHelpBroker.enableHelpKey(this.getRootPane(), m_sHelpID, null);

      Behavior oBehavior = null;
      int i, j;
      //Find our chosen behavior
      for (i = 0; i < mp_oBehaviors.length; i++) {
        for (j = 0; j < mp_oBehaviors[i].length; j++) {
          if (mp_oBehaviors[i][j].GetDescriptor().equals(sBehaviorName)) {
            oBehavior = mp_oBehaviors[i][j];
          }
        }
      }

      //If this behavior cannot be edited, set up a messagebox saying so
      if (oBehavior.m_bCanEditAutomatically == false) {
        JLabel jLabel = new JLabel("Please use this behavior's own dialog to " +
                                   "edit its settings.");
        jLabel.setFont(new ModelFont());
        jLabel.setAlignmentX(JComponent.CENTER_ALIGNMENT);

        JButton jOKButton = new JButton("OK");
        jOKButton.setFont(new ModelFont());
        jOKButton.setActionCommand("Cancel"); //the "Cancel" is on purpose
        jOKButton.addActionListener(this);
        jOKButton.setAlignmentX(JComponent.CENTER_ALIGNMENT);

        getContentPane().setLayout(new BoxLayout(getContentPane(),
                                                 BoxLayout.PAGE_AXIS));
        getContentPane().add(jLabel);
        getContentPane().add(jOKButton);

        return;
      }

      m_sBehaviorName = sBehaviorName;

      /********************************************************************
       * Leftmost panel - two lists with species and type for choosing
       ********************************************************************/
      JPanel jNewComboPicker = new JPanel();
      //Species choices
      String[] p_sSpeciesOptions = new String[m_oPop.GetNumberOfSpecies()];
      for (i = 0; i < p_sSpeciesOptions.length; i++) {
        p_sSpeciesOptions[i] = m_oPop.GetSpeciesNameFromCode(i).replace('_',
            ' ');
      }

      //Type choices
      String[] p_sTypeOptions = new String[m_oPop.GetNumberOfTypes()];
      for (i = 0; i < p_sTypeOptions.length; i++) {
        p_sTypeOptions[i] = m_oPop.GetTypeNameFromCode(i);
      }

      m_jSpecies = new JList(p_sSpeciesOptions);
      m_jTypes = new JList(p_sTypeOptions);
      m_jSpecies.setFont(new ModelFont());
      m_jTypes.setFont(new ModelFont());
      m_jSpecies.setVisibleRowCount(6);
      m_jTypes.setVisibleRowCount(6);
      JScrollPane jSpeciesScroller = new JScrollPane(m_jSpecies);
      JScrollPane jTypeScroller = new JScrollPane(m_jTypes);

      //Add button
      JButton jAddButton = new JButton(new ModelIcon(15, 15,
          ModelIcon.RIGHT_TRIANGLE));
      jAddButton.setFont(new ModelFont());
      jAddButton.setToolTipText(
          "Add this tree species and type to assigned list");
      jAddButton.setActionCommand("Add");
      jAddButton.addActionListener(this);

      //Assemble the panel
      jNewComboPicker.setLayout(new BoxLayout(jNewComboPicker,
                                              BoxLayout.PAGE_AXIS));
      JLabel jLabel = new JLabel("Please choose the species:");
      jLabel.setFont(new ModelFont());
      jNewComboPicker.add(jLabel);
      jNewComboPicker.add(jSpeciesScroller);
      jLabel = new JLabel("Please choose the tree types:");
      jLabel.setFont(new ModelFont());
      jNewComboPicker.add(jLabel);
      jNewComboPicker.add(jTypeScroller);
      //jNewComboPicker.add(jAddButton);

      /********************************************************************
       * List showing existing species/type combos
       ********************************************************************/
      JPanel jExisting = new JPanel(new java.awt.BorderLayout());
      m_jAssignedCombosListModel = new DefaultListModel();
      m_jAssignedCombosList = new JList(m_jAssignedCombosListModel);
      m_jAssignedCombosList.setFont(new ModelFont());
      JScrollPane jScroller = new JScrollPane(m_jAssignedCombosList);
      jScroller.setPreferredSize(new Dimension(200,
                                               (int) jScroller.getPreferredSize().
                                               getHeight()));

      //Load the combos
      for (i = 0; i < oBehavior.GetNumberOfCombos(); i++) {
        m_jAssignedCombosListModel.addElement(
            new ComboDisplay(oBehavior.GetSpeciesTypeCombo(i).GetSpecies(),
                             oBehavior.GetSpeciesTypeCombo(i).GetType()));
      }
      jLabel = new JLabel("Behavior currently assigned to:");
      jLabel.setFont(new ModelFont());
      jExisting.add(jLabel, java.awt.BorderLayout.NORTH);
      jExisting.add(jScroller, java.awt.BorderLayout.CENTER);

      /********************************************************************
       * Remove button for combos
       ********************************************************************/
      JButton jRemoveButton = new JButton("Remove");
      jRemoveButton.setFont(new ModelFont());
      jRemoveButton.setToolTipText(
          "Remove this tree species / type from this behavior's list");
      jRemoveButton.setActionCommand("Remove");
      jRemoveButton.addActionListener(this);

      //Assemble the entire window
      //Package everything but the buttons in a center panel
      JPanel jCenterPanel = new JPanel();
      jCenterPanel.setLayout(new FlowLayout());
      jCenterPanel.add(jNewComboPicker);
      jCenterPanel.add(jAddButton);
      jCenterPanel.add(jExisting);
      jCenterPanel.add(jRemoveButton);

      getContentPane().setLayout(new BorderLayout());
      getContentPane().add(jCenterPanel, BorderLayout.CENTER);
      //Buttons at the button
      getContentPane().add(new OKCancelButtonPanel(this, oHelpBroker, m_sHelpID),
                           BorderLayout.PAGE_END);
    }

    /**
     * Performs actions for clicked buttons.
     * @param e ActionEvent object.
     *
     * <br>Edit history:
     * <br>------------------
     * <br>April 28, 2004: Submitted in beta version (LEM)
     */
    public void actionPerformed(ActionEvent e) {
      if (e.getActionCommand().equals("Cancel")) {
        this.setVisible(false);
        this.dispose();
      }
      else if (e.getActionCommand().equals("OK")) {
        try {
          //Sets the combos listed
          //Find the behavior to which this applies
          ComboDisplay oCombo;
          Behavior oBehavior = null;
          int i, j;
          for (i = 0; i < mp_oBehaviors.length; i++) {
            for (j = 0; j < mp_oBehaviors[i].length; j++) {
              if (m_sBehaviorName.equals(mp_oBehaviors[i][j].GetDescriptor())) {
                oBehavior = mp_oBehaviors[i][j];
              }
            }
          }

          oBehavior.ClearSpeciesTypeCombos();
          for (i = 0; i < m_jAssignedCombosListModel.size(); i++) {
            oCombo = (ComboDisplay) m_jAssignedCombosListModel.elementAt(i);
            oBehavior.AddSpeciesTypeCombo(new SpeciesTypeCombo(oCombo.
                GetSpecies(), oCombo.GetType(), m_oPop));
          }

          this.setVisible(false);
          this.dispose();
        }
        catch (ModelException oErr) {
          ErrorGUI oHandler = new ErrorGUI(this);
          oHandler.WriteErrorMessage(oErr);
        }
      }
      else if (e.getActionCommand().equals("Add")) {

        //Adding a new type/species combo
        Object[] p_oSelectedSpecies, p_oSelectedTypes;
        ComboDisplay oCombo;
        String sSpecies, sType;
        int iSpecies, iType, i, j, k;
        boolean bDuplicate;

        //Get the species
        p_oSelectedSpecies = m_jSpecies.getSelectedValues();
        if (p_oSelectedSpecies == null || p_oSelectedSpecies.length == 0) {
          JOptionPane.showMessageDialog(this, "Please select a tree species.");
          return; //no species picked
        }

        //Get the types
        p_oSelectedTypes = m_jTypes.getSelectedValues();
        if (p_oSelectedTypes == null || p_oSelectedTypes.length == 0) {
          JOptionPane.showMessageDialog(this, "Please select a tree type.");
          return; //no types picked
        }

        for (i = 0; i < p_oSelectedSpecies.length; i++) {
          sSpecies = (String) p_oSelectedSpecies[i];
          iSpecies = m_oPop.GetSpeciesCodeFromName(sSpecies.replace(' ', '_'));
          for (j = 0; j < p_oSelectedTypes.length; j++) {
            bDuplicate = false;

            //Make sure this species/type combo doesn't already exist
            sType = (String) p_oSelectedTypes[j];
            iType = m_oPop.GetTypeCodeFromName(sType);

            for (k = 0; k < m_jAssignedCombosListModel.size(); k++) {
              oCombo = (ComboDisplay) m_jAssignedCombosListModel.elementAt(k);
              if (iSpecies == oCombo.GetSpecies() && iType == oCombo.GetType()) {
                bDuplicate = true;
              }
            }

            //Add the new combo
            if (bDuplicate == false) {
              oCombo = new ComboDisplay(iSpecies, iType);
              m_jAssignedCombosListModel.addElement(oCombo);
            }
          }
        }
      }
      else if (e.getActionCommand().equals("Remove")) {

        //Removing one or more type/species combos
        Object[] p_oSelected = m_jAssignedCombosList.getSelectedValues();
        if (p_oSelected == null || p_oSelected.length == 0) {
          JOptionPane.showMessageDialog(this, "Please select a tree type.");
          return; //Nothing selected
        }
        for (int i = 0; i < p_oSelected.length; i++) {
          m_jAssignedCombosListModel.removeElement(p_oSelected[i]);
        }
      }
    }
  }

  /**
   * Combines a text description of species/type combos with the actual data,
   * for display in lists.
   * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
   * <p>Company: Institute of Ecosystem Studies</p>
   * @author Lora E. Murphy
   * @version 1.0
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  private class ComboDisplay {
    private String m_sDescriptor;
    private int
        /**Combo species*/
        m_iSpecies,
        /**Combo type*/
        m_iType;

    /**
     * Constructor.  Formats a display string.
     * @param iSpecies Species for this combo.
     * @param iType Type for this combo.
     *
     * <br>Edit history:
     * <br>------------------
     * <br>April 28, 2004: Submitted in beta version (LEM)
     */
    public ComboDisplay(int iSpecies, int iType) {
      m_iSpecies = iSpecies;
      m_iType = iType;
      if (m_iType != TreePopulation.WOODY_DEBRIS) {
        m_sDescriptor = m_oPop.GetSpeciesNameFromCode(iSpecies).replace('_',
            ' ') + " " + m_oPop.GetTypeNameFromCode(iType) + "s";
      }
      else {
        m_sDescriptor = m_oPop.GetSpeciesNameFromCode(iSpecies).replace('_',
            ' ') + " " + m_oPop.GetTypeNameFromCode(iType);
      }
    }

    /**
     * Overridden so that this can be displayed in JLists.
     * @return String description
     *
     * <br>Edit history:
     * <br>------------------
     * <br>April 28, 2004: Submitted in beta version (LEM)
     */
    public String toString() {
      return m_sDescriptor;
    }

    /**
     * Gets the species.
     * @return The species number.
     *
     * <br>Edit history:
     * <br>------------------
     * <br>April 28, 2004: Submitted in beta version (LEM)
     */
    public int GetSpecies() {
      return m_iSpecies;
    }

    /**
     * Gets the type.
     * @return The type number.
     *
     * <br>Edit history:
     * <br>------------------
     * <br>April 28, 2004: Submitted in beta version (LEM)
     */
    public int GetType() {
      return m_iType;
    }
  }

  /**
   * Provides multi-line text wrapping to our list boxes
   * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
   * <p>Company: Institute of Ecosystem Studies</p>
   * @author Lora E. Murphy
   * @version 1.0
   */
  class ListRenderer
      extends MultilineLabel
      implements ListCellRenderer {

    /**The width of the component in which this will be displayed.  This is
     * float to save a casting step later when rendering.*/
    private float m_fComponentWidth;

    /**
     * Constructor.
     * @param iWidth Width of the component in which this renderer will display
     * text.
     */
    public ListRenderer(int iWidth) {
      m_fComponentWidth = iWidth;
    }

    /*
     * This method finds the image and text corresponding
     * to the selected value and returns the label, set up
     * to display the text and image.
     */
    public Component getListCellRendererComponent(
        JList list,
        Object value,
        int index,
        boolean isSelected,
        boolean cellHasFocus) {

      String sVal = String.valueOf(value);

      this.setText(sVal);
      if (isSelected) {

        setBackground(list.getSelectionBackground());
        setForeground(list.getSelectionForeground());
      }
      else {
        if (index % 2 == 1) {
          setBackground(new java.awt.Color(225, 225, 225));
        }
        else {
          setBackground(list.getBackground());
        }
        setForeground(list.getForeground());
      }

      //Make the label display enough rows to show all the text
      JLabel jNotUsedField = new JLabel(sVal);
      jNotUsedField.setFont(new ModelFont());
      float fWidth = (float) jNotUsedField.getPreferredSize().getWidth() - 15;
      int iNumRows = (int) java.lang.Math.ceil(fWidth / m_fComponentWidth);
      setRows(iNumRows);

      return this;
    }
  }
}
