package javawrapper;

import java.util.Vector;
import java.io.BufferedWriter;
import org.xml.sax.Attributes;

/**
 * This class contains methods common to all objects that contain data.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 * <br>July 19, 2004:  Added help ID (LEM)
 */
abstract public class WorkerBase {
  /**GUIManager object*/
  protected GUIManager m_oManager;

  /**User-consumable name for this object*/
  protected String m_sName;

  /**Highest-level XML tag for this object*/
  protected String m_sXMLTag;

  /**ID of help topic*/
  protected String m_sHelpID = "";

  /**All data for this object.  It should be placed in the order
   * in which it should be written in XML.*/
  protected Vector mp_oAllData;

  /**Whether or not to auto-display this object's parameters.*/
  protected boolean m_bAutoDisplayParameters;

  /**
   * Constructor.
   * @param oManager GUIManager object
   * @param sName User-consumable name for this object.
   * @param sXMLTag Parent XML tag for all of this object's data
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  WorkerBase(GUIManager oManager, String sName, String sXMLTag) {
    m_oManager = oManager;
    m_sName = sName;
    m_sXMLTag = sXMLTag;
    mp_oAllData = new Vector();
    m_bAutoDisplayParameters = true;
  }

  /**
   * Returns the user-acceptable name for this object.
   * @return Name.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public String GetName() {
    return m_sName;
  }

  /**
   * Does any needed setup once the tree population object has data.  This will
   * be called once the tree population has its data.
   * @param oPop TreePopulation object.
   * @throws ModelException if there are problems with setup.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  abstract public void DoSetup(TreePopulation oPop) throws ModelException;

  /**
   * Gets whether to auto-display this object's parameters.
   * @return Whether to auto-display parameters.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean GetAutoDisplayParameters() {
    return m_bAutoDisplayParameters;
  }

  /**
   * Gets this object's help ID, if it exists.
   * @return The help ID, or "" if there is no help ID.
   */
  public String GetHelpID() {
    return m_sHelpID;
  }

  /**
   * Formats data for display in a set of JTables.  For all required behaviors,
   * their required data objects are placed in a table.
   * Each element in the vector returned is one table to display, stored in a
   * vector.  Each single table vector's first element is an Object array with
   * the header row for the table, and the second element is an Object[][]
   * with the table's data.  Data labels should be placed in the first array
   * bucket.
   *
   * This function operates by looking at each behavior.  For each behavior
   * which is enabled, its required data objects are added to a data vector.
   * The vector is a vector of vectors where the data objects are grouped
   * together by number of columns they require.
   *
   * Once the sorted lists are in place, they are transformed to the table
   * format and returned.
   * @param oPop TreePopulation object.
   * @return Vector of vectors suitable for creating a set of JTables, or null
   * if there is no data to display.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>October 7, 2004: Integrated the BehaviorTypeBase version of this back
   * into here, and made it aware of the ModelVector "must apply to all
   * species" member
   */
  public Vector FormatDataForDisplay(TreePopulation oPop) {

    Vector p_oSingles = new Vector(), //single data objects
        p_oSpeciesSpecific = new Vector(); //SpeciesSpecific objects - vector of vectors
    //grouped by what species they apply to
    ModelData p_oDataObject;

    boolean[] p_bAppliesTo = null,
        p_bAppliesToAll = new boolean[oPop.GetNumberOfSpecies()];
    int i, j;
    boolean bAny = false;

    for (i = 0; i < p_bAppliesToAll.length; i++) {
      p_bAppliesToAll[i] = true;
    }

    if (this instanceof BehaviorTypeBase) {
      //First check to see if any child behaviors are enabled - if not, return
      //null
      BehaviorTypeBase oBase = (BehaviorTypeBase)this;
      for (i = 0; i < oBase.mp_oChildBehaviors.length; i++) {
        if (oBase.mp_oChildBehaviors[i].m_bIsEnabled) {
          bAny = true;
          break;
        }
      }
      if (!bAny) {
        return null;
      }

      //Go through the child behaviors again and start adding and sorting them
      for (i = 0; i < oBase.mp_oChildBehaviors.length; i++) {
        if (oBase.mp_oChildBehaviors[i].m_bIsEnabled) {
          p_bAppliesTo = oBase.mp_oChildBehaviors[i].GetWhichSpeciesUsed(oPop);
          for (j = 0;
               j < oBase.mp_oChildBehaviors[i].GetNumberOfRequiredDataObjects();
               j++) {

            p_oDataObject = (ModelData) oBase.mp_oChildBehaviors[i].
                GetRequiredData(j);
            if (p_oDataObject instanceof ModelVector &&
                ( (ModelVector) p_oDataObject).GetMustApplyToAllSpecies()) {
              AddDataObjectToDisplayArrays(p_oDataObject, p_oSingles,
                                           p_oSpeciesSpecific, p_bAppliesToAll);
            }
            else {
              AddDataObjectToDisplayArrays(p_oDataObject, p_oSingles,
                                           p_oSpeciesSpecific, p_bAppliesTo);
            }
          }
        }
      }
    }
    else {
      for (i = 0; i < mp_oAllData.size(); i++) {
        p_oDataObject = (ModelData) mp_oAllData.elementAt(i);
        if (p_oDataObject instanceof ModelVector &&
            ( (ModelVector) p_oDataObject).GetMustApplyToAllSpecies()) {
          AddDataObjectToDisplayArrays(p_oDataObject, p_oSingles,
                                       p_oSpeciesSpecific, p_bAppliesToAll);
        }
        else {
          AddDataObjectToDisplayArrays(p_oDataObject, p_oSingles,
                                       p_oSpeciesSpecific, p_bAppliesTo);
        }
      }
    }

    return FormatTable(p_oSingles, p_oSpeciesSpecific, oPop);
  }

  /**
   * Writes the XML data to a parameter file for the behaviors owned by this
   * object.
   * @param jOut File stream to write to.
   * @param oPop TreePopulation object.
   * @throws ModelException if there is a problem writing the file or
   * validating the data.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void WriteXML(BufferedWriter jOut, TreePopulation oPop) throws
      ModelException {
    try {

      jOut.write("<" + m_sXMLTag + ">");
      Object oData;
      ModelData oDataPiece;
      ModelVector p_oDataVector;
      int i;
      for (i = 0; i < mp_oAllData.size(); i++) {
        oData = mp_oAllData.elementAt(i);
        if (IsThisDataUsed( (ModelData) oData)) {
          if (oData instanceof ModelVector) {
            p_oDataVector = (ModelVector) oData;
            WriteSpeciesSpecificValue(jOut, p_oDataVector, oPop);
          }
          else if (oData instanceof ModelData) {
            oDataPiece = (ModelData) oData;
            WriteData(jOut, oDataPiece);
          }
        }
      }
      jOut.write("</" + m_sXMLTag + ">");
    }
    catch (java.io.IOException oExp) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.BAD_FILE;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = "There was a problem writing the parameter file.";
      throw (oErr);
    }
  }

  /**
   * Whether this piece of data is used and should be written to an XML file.
   * Override this if all data is not to be used.
   * @param oData Data to check.
   * @return Always true.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean IsThisDataUsed(ModelData oData) {
    return true;
  }

  /**
   * Sets the values for vector variables.  This method replaces values in
   * the vector with the values passed, in the order in which they are in
   * the array.  Only those vector indexes for which p_bApplies = true at that
   * same index are set.
   * @param oData Vector data variable to set.
   * @param p_fValuesToSet Array of values to put into the vector.
   * @param p_bAppliesTo Array of flags for which values should be set.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetVectorValues(ModelVector oData, Float[] p_fValuesToSet,
                              boolean[] p_bAppliesTo) {
    int i;

    oData.GetValue().setSize(p_fValuesToSet.length);

    for (i = 0; i < p_fValuesToSet.length; i++) {
      if (p_bAppliesTo[i]) {
        oData.GetValue().remove(i);
        oData.GetValue().add(i, new Float(p_fValuesToSet[i].floatValue()));
      }
    }
  }

  /**
   * Sets the values for vector variables.  This method replaces values in
   * the vector with the values passed, in the order in which they are in
   * the array.
   * @param oData Vector data variable to set.
   * @param p_fValuesToSet Array of values to put into the vector.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetVectorValues(ModelVector oData, Float[] p_fValuesToSet) {
    boolean[] p_bAppliesTo = new boolean[p_fValuesToSet.length];
    int i;
    for (i = 0; i < p_bAppliesTo.length; i++) {
      p_bAppliesTo[i] = true;
    }
    SetVectorValues(oData, p_fValuesToSet, p_bAppliesTo);
  }

  /**
   * Sets the values for vector variables.  This method replaces all values in
   * the vector with the values passed, in the order in which they are in
   * the array.
   * @param oData Vector data variable to set.
   * @param p_iValuesToSet Array of values to put into the vector.
   * @param p_bAppliesTo Array of flags for which values should be set.
   * @todo Make this reject the setting of values for non-int vectors.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetVectorValues(ModelVector oData, Integer[] p_iValuesToSet,
                              boolean[] p_bAppliesTo) {
    int i;
    oData.GetValue().setSize(p_iValuesToSet.length);

    for (i = 0; i < p_iValuesToSet.length; i++) {
      if (p_bAppliesTo[i]) {
        oData.GetValue().remove(i);
        oData.GetValue().add(i, new Integer(p_iValuesToSet[i].intValue()));
      }
    }
  }

  /**
   * Sets the values for vector variables.  This method replaces all values in
   * the vector with the values passed, in the order in which they are in
   * the array.
   * @param oData Vector data variable to set.
   * @param p_iValuesToSet Array of values to put into the vector.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetVectorValues(ModelVector oData, Integer[] p_iValuesToSet) {
    boolean[] p_bAppliesTo = new boolean[p_iValuesToSet.length];
    int i;
    for (i = 0; i < p_bAppliesTo.length; i++) {
      p_bAppliesTo[i] = true;
    }
    SetVectorValues(oData, p_iValuesToSet, p_bAppliesTo);
  }

  /**
   * Sets the values for vector variables.  The values in the vector which
   * contains the values being assigned are cast to match the data types in
   * the vector to which the values are being assigned.  This method replaces
   * all values in the vector with the values passed, in the order in which
   * they are in the array.
   * @param oData Vector data variable to set.
   * @param p_oValuesToSet Vector of values to put into the vector.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
//  public void SetVectorValues(ModelVector oData, Vector p_oValuesToSet) {

  // }

  /**
   * Sets the values for vector variables.  The values in the vector which
   * contains the values being assigned are cast to match the data types in
   * the vector to which the values are being assigned.  Only those vector
   * indexes for which p_bApplies = true at that same index are set.
   * @param p_oData Vector data variable to set.
   * @param p_oValuesToSet Vector of values to put into the vector.
   * @param p_bAppliesTo Array of flags for which values should be set.
   * @throws ModelException if the values cannot be cast to the correct type
   * and assigned.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetVectorValues(ModelVector p_oData, Vector p_oValuesToSet,
                              boolean[] p_bAppliesTo) throws ModelException {

    try {
      int i;
      //Determine what kind of data the vector wants
      if (p_oData.GetDataType() == ModelVector.INTEGER) {

        //Try to make the vector of values to set into an array of Integers
        Integer[] p_oInts = new Integer[p_oValuesToSet.size()];
        for (i = 0; i < p_oValuesToSet.size(); i++) {
          if (p_oValuesToSet.elementAt(i) instanceof String) {
            String sVal = (String) p_oValuesToSet.elementAt(i);
            if (sVal.length() == 0) {
              p_oInts[i] = new Integer(0);
            }
            else {
              p_oInts[i] = new Integer(sVal);
            }
          }
          else if (null == p_oValuesToSet.elementAt(i)) {
            p_oInts[i] = new Integer(0);
          }
          else {
            p_oInts[i] = (Integer) p_oValuesToSet.elementAt(i);
          }
        }

        SetVectorValues(p_oData, p_oInts, p_bAppliesTo);

      }
      else if (p_oData.GetDataType() == ModelVector.FLOAT) {
        //Try to make the vector of values to set into an array of Floats
        Float[] p_oFloats = new Float[p_oValuesToSet.size()];
        for (i = 0; i < p_oFloats.length; i++) {
          if (p_oValuesToSet.elementAt(i) instanceof String) {
            String sVal = (String) p_oValuesToSet.elementAt(i);
            if (sVal.length() == 0) {
              p_oFloats[i] = new Float(0);
            }
            else {
              p_oFloats[i] = new Float(sVal);
            }
          }
          else if (null == p_oValuesToSet.elementAt(i)) {
            p_oFloats[i] = new Float(0);
          }
          else {
            p_oFloats[i] = (Float) p_oValuesToSet.elementAt(i);
          }
        }

        SetVectorValues(p_oData, p_oFloats, p_bAppliesTo);
      }
      else if (p_oData.GetDataType() == ModelVector.MODEL_ENUM) {
        ModelEnum oData;

        p_oData.GetValue().setSize(p_oValuesToSet.size());

        for (i = 0; i < p_oValuesToSet.size(); i++) {
          if (p_bAppliesTo[i]) {
            oData = (ModelEnum) p_oData.GetValue().elementAt(i);
            if (p_oValuesToSet.elementAt(i) instanceof String) {
              //Try to make the string into an integer - if it doesn't work,
              //set the string directly
              try {
                Integer oInt = new Integer( (String) p_oValuesToSet.elementAt(i));
                oData.SetValue(oInt.intValue());
              }
              catch (java.lang.NumberFormatException oE) {
                oData.SetValue( (String) p_oValuesToSet.elementAt(i));
              }
            }
          }
        }
      }
      else {
        //Panic
        ModelException oErr = new ModelException();
        oErr.sMessage = "For object " + m_sName + ", data " +
            p_oData.GetDescriptor() + ", can't determine vector data type.";
        oErr.sCallingFunction = "JAVA";
        throw (oErr);
      }
    }
    catch (java.lang.ClassCastException oErr) {
      ModelException oExp = new ModelException();
      if (p_oData != null) {
        oExp.sMessage = "Attempt to set " + p_oData.GetDescriptor() +
            " with the wrong data type.";
      }
      oExp.sCallingFunction = "JAVA";
      throw (oExp);
    }
    catch (java.lang.NumberFormatException oErr) {
      ModelException oExp = new ModelException();
      if (p_oData != null) {
        oExp.sMessage = "Attempt to set " + p_oData.GetDescriptor() +
            " with the wrong data type.";
      }
      oExp.sCallingFunction = "JAVA";
      throw (oExp);
    }

  }

  /**
   * Makes sure the elements of a vector are positive numbers (greater than
   * zero).  Only those indexes for which p_bAppliesTo = true at
   * the same index are checked.
   * @param oData The vector to check.
   * @param p_bAppliesTo Set of flags for which values to check
   * @throws ModelException if there are any non-positive number values in the
   * vector.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void MakeSureAllPositive(ModelVector oData, boolean[] p_bAppliesTo) throws
      ModelException {
    int i;
    Number oNumber;
    for (i = 0; i < oData.GetValue().size(); i++) {
      if (p_bAppliesTo[i]) {
        Object oObject = oData.GetValue().elementAt(i);
        if (oObject instanceof Number) {
          oNumber = (Number) oData.GetValue().elementAt(i);
          if (oNumber == null || oNumber.floatValue() <= 0) {
            ModelException oErr = new ModelException();
            oErr.iErrorCode = ErrorGUI.BAD_DATA;
            oErr.sCallingFunction = "JAVA";
            oErr.sMessage = m_sName + ":  All values in " + oData.GetDescriptor() +
                " must be greater than zero.";
            throw (oErr);
          }
        }
        else {
          ModelException oErr = new ModelException(ErrorGUI.BAD_DATA,
              "JAVA",
              m_sName + ":  All values in " + oData.GetDescriptor() +
              " must be numbers.");
          throw (oErr);

        }
      }
    }
  }

  /**
   * Makes sure the elements of a vector are positive numbers (greater than
   * zero).
   * @param oData The vector to check.
   * @throws ModelException if there are any non-positive number values in the
   * vector.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void MakeSureAllPositive(ModelVector oData) throws
      ModelException {
    boolean[] p_bAppliesTo = new boolean[oData.GetValue().size()];
    int i;
    for (i = 0; i < p_bAppliesTo.length; i++) {
      p_bAppliesTo[i] = true;
    }
    MakeSureAllPositive(oData, p_bAppliesTo);
  }

  /**
   * Makes sure the elements of a vector are non-negative numbers (greater
   * than or equal to zero).  Only those indexes for which p_bAppliesTo = true
   * at the same index are checked.
   * @param oData The vector to check.
   * @param p_bAppliesTo Set of flags for which values to check
   * @throws ModelException if there are any negative number values in the
   * vector.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void MakeSureAllNonNegative(ModelVector oData, boolean[] p_bAppliesTo) throws
      ModelException {
    int i;
    Number oNumber;
    for (i = 0; i < oData.GetValue().size(); i++) {
      if (p_bAppliesTo[i]) {
        oNumber = (Number) oData.GetValue().elementAt(i);
        if (oNumber == null || oNumber.floatValue() < 0) {
          ModelException oErr = new ModelException();
          oErr.iErrorCode = ErrorGUI.BAD_DATA;
          oErr.sCallingFunction = "JAVA";
          oErr.sMessage = m_sName + ":  All values in " + oData.GetDescriptor() +
              " must be greater than or equal to zero.";
          throw (oErr);
        }
      }
    }
  }

  /**
   * Makes sure the elements of a vector are non-negative numbers (greater
   * than or equal to zero).
   * @param oData The vector to check.
   * @throws ModelException if there are any negative number values in the
   * vector.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void MakeSureAllNonNegative(ModelVector oData) throws
      ModelException {
    boolean[] p_bAppliesTo = new boolean[oData.GetValue().size()];
    int i;
    for (i = 0; i < p_bAppliesTo.length; i++) {
      p_bAppliesTo[i] = true;
    }
    MakeSureAllNonNegative(oData, p_bAppliesTo);
  }

  /**
   * Makes sure the elements of a vector are non-zero.
   * @param oData The vector to check.
   * @throws ModelException if there are any 0 values in the
   * vector.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>September 6, 2004: Created (LEM)
   */
  public void MakeSureAllNonZero(ModelVector oData) throws
      ModelException {
    boolean[] p_bAppliesTo = new boolean[oData.GetValue().size()];
    int i;
    for (i = 0; i < p_bAppliesTo.length; i++) {
      p_bAppliesTo[i] = true;
    }
    MakeSureAllNonZero(oData, p_bAppliesTo);
  }

  /**
   * Makes sure the elements of a vector are non-zero numbers.  Only those
   * indexes for which p_bAppliesTo = true at the same index are checked.
   * @param oData The vector to check.
   * @param p_bAppliesTo Set of flags for which values to check
   * @throws ModelException if there are any negative number values in the
   * vector.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>September 6, 2004: Created (LEM)
   */
  public void MakeSureAllNonZero(ModelVector oData, boolean[] p_bAppliesTo) throws
      ModelException {
    int i;
    Number oNumber;
    for (i = 0; i < oData.GetValue().size(); i++) {
      if (p_bAppliesTo[i]) {
        oNumber = (Number) oData.GetValue().elementAt(i);
        if (oNumber == null || oNumber.floatValue() == 0) {
          ModelException oErr = new ModelException();
          oErr.iErrorCode = ErrorGUI.BAD_DATA;
          oErr.sCallingFunction = "JAVA";
          oErr.sMessage = m_sName + ":  No values in " + oData.GetDescriptor() +
              " may equal zero.";
          throw (oErr);
        }
      }
    }
  }

  /**
   * Makes sure that all the values in a vector fall between two bounds.  Only
   * those indexes for which p_bAppliesTo = true at the same index are checked.
   * @param oData ModelVector Vector to check.
   * @param p_bAppliesTo boolean[] Set of flags for which values to check.
   * @param fLowerBound float Minimum acceptable value.
   * @param fUpperBound float Maximum acceptable value.
   * @throws ModelException If any value falls outside of the range.
   */
  public void MakeSureAllAreBounded(ModelVector oData, boolean[] p_bAppliesTo, float fLowerBound, float fUpperBound) throws ModelException {
    int i;
    Number oNumber;
    for (i = 0; i < oData.GetValue().size(); i++) {
      if (p_bAppliesTo[i]) {
        oNumber = (Number) oData.GetValue().elementAt(i);
        if (oNumber == null || oNumber.floatValue() < fLowerBound ||
            oNumber.floatValue() > fUpperBound) {
          ModelException oErr = new ModelException();
          oErr.iErrorCode = ErrorGUI.BAD_DATA;
          oErr.sCallingFunction = "JAVA";
          oErr.sMessage = m_sName + ":  All values in " + oData.GetDescriptor() +
              " must be between " + fLowerBound + " and " + fUpperBound + ".";
          throw (oErr);
        }
      }
    }
  }

  /**
   * Makes sure the elements of a float array are positive numbers (greater
   * than zero).  Only those indexes for which p_bAppliesTo = true at
   * the same index are checked.
   * @param p_fData The array to check.
   * @param sName The name of the vector, to put in the error message.
   * @param p_bAppliesTo Set of flags for which values to check
   * @throws ModelException if there are any non-positive number values in the
   * vector.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void MakeSureAllPositive(Float[] p_fData, String sName,
                                  boolean[] p_bAppliesTo) throws
      ModelException {
    int i;
    for (i = 0; i < p_fData.length; i++) {
      if (p_bAppliesTo[i]) {
        if (p_fData[i] == null || p_fData[i].compareTo(new Float(0)) <= 0) {
          ModelException oErr = new ModelException();
          oErr.iErrorCode = ErrorGUI.BAD_DATA;
          oErr.sCallingFunction = "JAVA";
          oErr.sMessage = m_sName + ":  All values in " + sName +
              " must be greater than zero.";
          throw (oErr);
        }
      }
    }
  }

  /**
   * Makes sure the elements of a float array are positive numbers (greater
   * than zero).
   * @param p_fData The array to check.
   * @param sName The name of the vector, to put in the error message.
   * @throws ModelException if there are any non-positive number values in the
   * vector.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void MakeSureAllPositive(Float[] p_fData, String sName) throws
      ModelException {
    boolean[] p_bAppliesTo = new boolean[p_fData.length];
    int i;
    for (i = 0; i < p_bAppliesTo.length; i++) {
      p_bAppliesTo[i] = true;
    }
    MakeSureAllPositive(p_fData, sName, p_bAppliesTo);
  }

  /**
   * Makes sure the elements of a float array are non-negative numbers
   * (greater than or equal to zero).  Only those indexes for which
   * p_bAppliesTo = true at the same index are checked.
   * @param p_fData The array to check.
   * @param sName The name of the vector, to put in the error message.
   * @param p_bAppliesTo Set of flags for which values to check
   * @throws ModelException if there are any negative number values in the
   * vector.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void MakeSureAllNonNegative(Float[] p_fData, String sName,
                                     boolean[] p_bAppliesTo) throws
      ModelException {
    int i;
    for (i = 0; i < p_fData.length; i++) {
      if (p_bAppliesTo[i]) {
        if (p_fData[i] == null || p_fData[i].compareTo(new Float(0)) < 0) {
          ModelException oErr = new ModelException();
          oErr.iErrorCode = ErrorGUI.BAD_DATA;
          oErr.sCallingFunction = "JAVA";
          oErr.sMessage = m_sName + ":  All values in " + sName +
              " must be greater than or equal to zero.";
          throw (oErr);
        }
      }
    }
  }

  /**
   * Makes sure the elements of a float array are non-negative numbers
   * (greater than or equal to zero).
   * @param p_fData The array to check.
   * @param sName The name of the vector, to put in the error message.
   * @throws ModelException if there are any negative number values in the
   * vector.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void MakeSureAllNonNegative(Float[] p_fData, String sName) throws
      ModelException {
    boolean[] p_bAppliesTo = new boolean[p_fData.length];
    int i;
    for (i = 0; i < p_bAppliesTo.length; i++) {
      p_bAppliesTo[i] = true;
    }
    MakeSureAllNonNegative(p_fData, sName, p_bAppliesTo);
  }

  /**
   * Makes sure the elements of an integer array are positive numbers
   * (greater than zero).  Only those indexes for which p_bAppliesTo = true at
   * the same index are checked.
   * @param p_iData The array to check.
   * @param sName The name of the vector, to put in the error message.
   * @param p_bAppliesTo Set of flags for which values to check
   * @throws ModelException if there are any non-positive number values in the
   * vector.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void MakeSureAllPositive(Integer[] p_iData, String sName,
                                  boolean[] p_bAppliesTo) throws
      ModelException {
    int i;
    for (i = 0; i < p_iData.length; i++) {
      if (p_bAppliesTo[i]) {
        if (p_iData[i] == null || p_iData[i].compareTo(new Integer(0)) <= 0) {
          ModelException oErr = new ModelException();
          oErr.iErrorCode = ErrorGUI.BAD_DATA;
          oErr.sCallingFunction = "JAVA";
          oErr.sMessage = m_sName + ":  All values in " + sName +
              " must be greater than zero.";
          throw (oErr);
        }
      }
    }
  }

  /**
   * Makes sure the elements of an integer array are positive numbers
   * (greater than zero).
   * @param p_iData The array to check.
   * @param sName The name of the vector, to put in the error message.
   * @throws ModelException if there are any non-positive number values in the
   * vector.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void MakeSureAllPositive(Integer[] p_iData, String sName) throws
      ModelException {
    boolean[] p_bAppliesTo = new boolean[p_iData.length];
    int i;
    for (i = 0; i < p_bAppliesTo.length; i++) {
      p_bAppliesTo[i] = true;
    }
    MakeSureAllPositive(p_iData, sName, p_bAppliesTo);
  }

  /**
   * Makes sure the elements of an integer array are non-negative numbers
   * (greater than or equal to zero).  Only those indexes for which
   * p_bAppliesTo = true at the same index are checked.
   * @param p_iData The array to check.
   * @param sName The name of the vector, to put in the error message.
   * @param p_bAppliesTo Set of flags for which values to check
   * @throws ModelException if there are any negative number values in the
   * vector.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void MakeSureAllNonNegative(Integer[] p_iData, String sName,
                                     boolean[] p_bAppliesTo) throws
      ModelException {
    int i;
    for (i = 0; i < p_iData.length; i++) {
      if (p_bAppliesTo[i]) {
        if (p_iData[i] == null || p_iData[i].compareTo(new Integer(0)) < 0) {
          ModelException oErr = new ModelException();
          oErr.iErrorCode = ErrorGUI.BAD_DATA;
          oErr.sCallingFunction = "JAVA";
          oErr.sMessage = m_sName + ":  All values in " + sName +
              " must be greater than or equal to zero.";
          throw (oErr);
        }
      }
    }
  }

  /**
   * Makes sure the elements of an integer array are non-negative numbers
   * (greater than or equal to zero).
   * @param p_iData The array to check.
   * @param sName The name of the vector, to put in the error message.
   * @throws ModelException if there are any negative number values in the
   * vector.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void MakeSureAllNonNegative(Integer[] p_iData, String sName) throws
      ModelException {
    boolean[] p_bAppliesTo = new boolean[p_iData.length];
    int i;
    for (i = 0; i < p_bAppliesTo.length; i++) {
      p_bAppliesTo[i] = true;
    }
    MakeSureAllNonNegative(p_iData, sName, p_bAppliesTo);
  }

  /**
   * Makes sure a value is greater than zero.
   * @param oValue The value to check.
   * @throws ModelException if the value is not greater than zero.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void MakeSurePositive(ModelFloat oValue) throws ModelException {
    if (oValue.GetValue() <= 0) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.BAD_DATA;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = m_sName + ":  " + oValue.GetDescriptor() +
          " must be greater than zero.";
      throw (oErr);
    }
  }

  /**
   * Makes sure a value is greater than or equal to zero.
   * @param oValue The value to check.
   * @throws ModelException if the value is not greater than zero.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void MakeSureNonNegative(ModelFloat oValue) throws ModelException {
    if (oValue.GetValue() < 0) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.BAD_DATA;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = m_sName + ":  " + oValue.GetDescriptor() +
          " must be greater than or equal to zero.";
      throw (oErr);
    }
  }

  /**
   * Makes sure a value is not equal to zero.
   * @param oValue The value to check.
   * @throws ModelException if the value is equal to zero (or in this case a
   * very small number).
   *
   * <br>Edit history:
   * <br>------------------
   * <br>December 8, 2005: Created (LEM)
   */
  public void MakeSureNonZero(ModelFloat oValue) throws ModelException {
    if (java.lang.Math.abs(oValue.GetValue()) < 1.0E-10) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.BAD_DATA;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = m_sName + ":  " + oValue.GetDescriptor() +
          " cannot be zero.";
      throw (oErr);
    }
  }

  /**
   * Makes sure a value is not equal to zero.
   * @param oValue The value to check.
   * @throws ModelException if the value is equal to zero.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>December 8, 2005: Created (LEM)
   */
  public void MakeSureNonZero(ModelInt oValue) throws ModelException {
    if (oValue.GetValue() == 0) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.BAD_DATA;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = m_sName + ":  " + oValue.GetDescriptor() + " cannot be zero.";
      throw (oErr);
    }
  }

  /**
   * Makes sure a value is greater than zero.
   * @param oValue The value to check.
   * @throws ModelException if the value is not greater than zero.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void MakeSurePositive(ModelInt oValue) throws ModelException {
    if (oValue.GetValue() <= 0) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.BAD_DATA;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = m_sName + ":  " + oValue.GetDescriptor() +
          " must be greater than zero.";
      throw (oErr);
    }
  }

  /**
   * Makes sure a value is greater than or equal to zero.
   * @param oValue The value to check.
   * @throws ModelException if the value is not greater than zero.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void MakeSureNonNegative(ModelInt oValue) throws ModelException {
    if (oValue.GetValue() < 0) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.BAD_DATA;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = m_sName + ":  " + oValue.GetDescriptor() +
          " must be greater than or equal to zero.";
      throw (oErr);
    }
  }

  /**
   * Makes sure a value is a proportion between 0 and 1 (inclusive).
   * @param fValue Value to check.
   * @throws ModelException if the value is not a valid proportion.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void MakeSureIsProportion(ModelFloat fValue) throws ModelException {
    if (fValue.GetValue() < 0 || fValue.GetValue() > 1) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.BAD_DATA;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = m_sName + ":  " + fValue.GetDescriptor() +
          " must be a value between 0 and 1.";
      throw (oErr);
    }
  }

  /**
   * Makes sure the data in a vector are proportions between 0
   * and 1 (inclusive).  Only those indexes for which p_bAppliesTo = true at
   * the same index are checked.
   * @param oData Vector to check
   * @param p_bAppliesTo Set of flags for which values to check
   * @throws ModelException if any value in the vector is not a valid
   * proportion.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void MakeSureAllAreProportions(ModelVector oData,
                                        boolean[] p_bAppliesTo) throws
      ModelException {
    int i;
    Float fNumber;
    for (i = 0; i < oData.GetValue().size(); i++) {
      if (p_bAppliesTo[i]) {
        fNumber = (Float) oData.GetValue().elementAt(i);
        if (fNumber == null || fNumber.floatValue() < 0 ||
            fNumber.floatValue() > 1) {
          ModelException oErr = new ModelException();
          oErr.iErrorCode = ErrorGUI.BAD_DATA;
          oErr.sCallingFunction = "JAVA";
          oErr.sMessage = m_sName + ":  All values in " + oData.GetDescriptor() +
              " must be values between 0 and 1.";
          throw (oErr);
        }
      }
    }
  }

  /**
   * Makes sure all the data in a vector are proportions between 0
   * and 1 (inclusive).
   * @param oData Vector to check
   * @throws ModelException if any value in the vector is not a valid
   * proportion.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void MakeSureAllAreProportions(ModelVector oData) throws
      ModelException {
    boolean[] p_bAppliesTo = new boolean[oData.GetValue().size()];
    int i;
    for (i = 0; i < p_bAppliesTo.length; i++) {
      p_bAppliesTo[i] = true;
    }
    MakeSureAllAreProportions(oData, p_bAppliesTo);
  }

  /**
   * Makes sure that a vector is of a certain size.
   * @param oData Vector to validate.
   * @param iLength Expected length.
   * @throws ModelException if the vector is not of the correct length.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void MakeSureRightSize(ModelVector oData, int iLength) throws
      ModelException {
    if (oData.GetValue().size() != iLength) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.BAD_DATA;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = m_sName + ":  There is the wrong number of values in " +
          oData.GetDescriptor() +
          ".  Expected: " + iLength + "; actual: " + oData.GetValue().size() +
          ".";
      throw (oErr);
    }
  }

  /**
   * Writes a set of species-specific float values to a parameter file.
   * @param jOut The file to write to.
   * @param p_oData The vector of data pieces.
   * @param oPop Tree population object to get species names
   * @throws java.io.IOException Passes on exceptions from FileWriter
   * @throws ModelException Passes on from TreePopulation
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void WriteSpeciesSpecificValue(BufferedWriter jOut, ModelVector p_oData,
                                        TreePopulation oPop) throws
      java.io.IOException, ModelException {

    if (p_oData == null || p_oData.GetValue().size() == 0) {
      return;
    }

    //Write an array that has every bucket set to true
    boolean[] p_bUse = new boolean[p_oData.GetValue().size()];
    int i;

    for (i = 0; i < p_bUse.length; i++) {
      p_bUse[i] = true;
    }

    WriteSpeciesSpecificValue(jOut, p_oData, oPop, p_bUse);
  }

  /**
   * Writes a set of species-specific values to a parameter file for a
   * certain subset of species.  If none of the species are enabled, nothing is
   * written.
   * @param jOut The file to write to.
   * @param p_oData The vector of data pieces.  It should be sized to equal the
   * total number of species, even if it doesn't have values for them all.
   * Each piece of data for a species should be at the element index
   * corresponding to the species number.
   * @param oPop Tree population object to get species names
   * @param p_bFlags An array, sized total number of species; for each species
   * number, whether or not to include data for the species in question.
   * @throws java.io.IOException Passes on exceptions from FileWriter
   * @throws ModelException Passes on from TreePopulation
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>October 7, 2004: Fixed this to work with the ModelVector "must apply
   * to all species" attribute (LEM)
   */
  public void WriteSpeciesSpecificValue(BufferedWriter jOut, ModelVector p_oData,
                                        TreePopulation oPop,
                                        boolean[] p_bFlags) throws
      java.io.IOException, ModelException {

    if (p_oData == null || p_oData.GetValue().size() == 0) {
      return;
    }

    boolean[] p_bFlagsCopy = new boolean[p_bFlags.length];
    int i;
    boolean bAnybodyUses = false;

    for (i = 0; i < p_bFlags.length; i++) {
      if (p_bFlags[i]) {
        bAnybodyUses = true;
      }
    }
    if (!bAnybodyUses) {
      return;
    }

    //Make a copy of the flags in case we need to override them
    for (i = 0; i < p_bFlags.length; i++) {
      p_bFlagsCopy[i] = p_bFlags[i];
    }

    if (p_oData.GetMustApplyToAllSpecies()) {
      for (i = 0; i < p_bFlagsCopy.length; i++) {
        p_bFlagsCopy[i] = true;
      }
    }

    jOut.write("<" + p_oData.GetXMLTag() + ">");

    for (i = 0; i < oPop.GetNumberOfSpecies(); i++) {
      if (p_bFlagsCopy[i]) {
        if (p_oData.GetValue().elementAt(i) != null) {
          jOut.write("<" + p_oData.GetChildXMLTag() + " species=\"" +
                     oPop.GetSpeciesNameFromCode(i).replace(' ', '_') +
                     "\">" + p_oData.GetValue().elementAt(i).toString() + "</" +
                     p_oData.GetChildXMLTag() + ">");
        }
      }
    }
    jOut.write("</" + p_oData.GetXMLTag() + ">");
  }

  /**
   * Writes a piece of data to an XML file.  If the string value is an empty
   * string, nothing is written.
   * @param ojOut The file to write to.
   * @param oData The data being written.
   * @throws java.io.IOException Passes on exceptions from FileWriter
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void WriteData(BufferedWriter ojOut, ModelData oData) throws java.io.
      IOException {
    String sValue = oData.toString().trim();
    if (sValue.length() > 0) {
      ojOut.write("<" + oData.GetXMLTag() + ">" + oData.toString() +
                  "</" + oData.GetXMLTag() + ">");
    }
  }

  /**
   * Formats and adds a value into an Object array suitable for placing in a
   * JTable.  The new array is formed by taking the descriptor and putting it
   * into the first position, and then putting the value in the next position.
   * @param p_oExisting The existing Object array to add to.
   * @param oToFormat The object to format.
   * @return Object array suitable for placing in a JTable.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected Object[][] FormatDataForTable(Object[][] p_oExisting,
                                          ModelFloat oToFormat) {
    if (null == oToFormat) {
      return p_oExisting;
    }

    //Create our new array
    Object[] p_oArray = new Object[2];
    p_oArray[0] = oToFormat.GetDescriptor();
    p_oArray[1] = new Float(oToFormat.GetValue());

    return AddDataToArray(p_oExisting, p_oArray);
  }

  /**
   * Formats and adds a value into an Object array suitable for placing in a
   * JTable.  The new array is formed by taking the descriptor and putting it
   * into the first position, and then putting the value in the next position.
   * @param p_oExisting The existing Object array to add to.
   * @param oToFormat The object to format.
   * @return Object array suitable for placing in a JTable.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected Object[][] FormatDataForTable(Object[][] p_oExisting,
                                          ModelInt oToFormat) {
    if (null == oToFormat) {
      return p_oExisting;
    }

    //Create our vectory
    Object[] p_oArray = new Object[2];
    p_oArray[0] = oToFormat.GetDescriptor();
    p_oArray[1] = new Integer(oToFormat.GetValue());

    return AddDataToArray(p_oExisting, p_oArray);
  }

  /**
   * Formats and adds a value into an Object array suitable for placing in a
   * JTable.  The new array is formed by taking the descriptor and putting it
   * into the first position, and then putting the value in the next position.
   * @param p_oExisting The existing Object array to add to.
   * @param oToFormat The object to format.
   * @return Object array suitable for placing in a JTable.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected Object[][] FormatDataForTable(Object[][] p_oExisting,
                                          ModelString oToFormat) {
    if (null == oToFormat) {
      return p_oExisting;
    }

    //Create our vectory
    Object[] p_oArray = new Object[2];
    p_oArray[0] = oToFormat.GetDescriptor();
    p_oArray[1] = oToFormat.GetValue();

    return AddDataToArray(p_oExisting, p_oArray);
  }

  /**
   * Formats a vector of values into an object array and adds it to another
   * Object array suitable for placing in a
   * JTable.  The descriptor is placed in the first bucket of the new vector
   * and all subsequent values in order after that.
   * @param p_oExisting The existing Object array to add to.
   * @param oToFormat The object to format.
   * @return Object array suitable for placing in a JTable.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected Object[][] FormatDataForTable(Object[][] p_oExisting,
                                          ModelVector oToFormat) {
    if (null == oToFormat) {
      return p_oExisting;
    }

    Object[] p_oArray = new Object[oToFormat.GetValue().size() + 1];
    int i;

    p_oArray[0] = oToFormat.GetDescriptor();
    for (i = 0; i < oToFormat.GetValue().size(); i++) {
      if (oToFormat.GetValue().elementAt(i) instanceof ModelEnum) {
        p_oArray[i +
            1] = FormatComboBoxString( (ModelEnum) oToFormat.GetValue().
                                      elementAt(i));
      }
      else {
        p_oArray[i + 1] = oToFormat.GetValue().elementAt(i);
      }
    }
    return AddDataToArray(p_oExisting, p_oArray);
  }

  /**
   * Formats a vector of values into a plain vector suitable for placing in a
   * JTable and adds it to a vector of vectors at the end.  No headers are
   * created.  The descriptor is placed in the first bucket of the new vector
   * and all subsequent values in order after that.
   * @param p_oExisting The existing Object array to add to.
   * @param oToFormat The object to format.
   * @param p_bWhichSpecies Boolean of whether or not each species should be
   * added.  If false for a species, the value at that Vector index is not
   * added.
   * @return Object array suitable for placing in a JTable.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected Object[][] FormatDataForTable(Object[][] p_oExisting,
                                          ModelVector oToFormat,
                                          boolean[] p_bWhichSpecies) {
    //If there's no vector to add, return the existing
    if (null == oToFormat) {
      return p_oExisting;
    }

    //Count the number of species used
    int i, iNumUsed = 0, iSize;
    for (i = 0; i < p_bWhichSpecies.length; i++) {
      if (p_bWhichSpecies[i]) {
        iNumUsed++;
      }
    }
    if (iNumUsed == 0) {
      return p_oExisting;
    }

    Object[] p_oArray = new Object[iNumUsed + 1];

    p_oArray[0] = oToFormat.GetDescriptor();
    iNumUsed = 1;
    iSize = oToFormat.GetValue().size();
    for (i = 0; i < iSize; i++) {
      if (p_bWhichSpecies[i]) {
        if (oToFormat.GetValue().elementAt(i) instanceof ModelEnum) {
          p_oArray[iNumUsed] = FormatComboBoxString( (ModelEnum) oToFormat.
              GetValue().elementAt(i));
        }
        else {
          p_oArray[iNumUsed] = oToFormat.GetValue().elementAt(i);
        }
        iNumUsed++;
      }
    }
    return AddDataToArray(p_oExisting, p_oArray);
  }

  /**
   * Adds an Object array to an existing Object array.  The new array
   * is added at the end.  If the existing array is null,
   * it will be created.
   * @param p_oExistingArray Object array to add to.
   * @param p_oAddArray Array to add.
   * @return The new Object array.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected Object[][] AddDataToArray(Object[][] p_oExistingArray,
                                      Object[] p_oAddArray) {

    if (null == p_oAddArray) {
      return p_oExistingArray;
    }

    //Is the existing array null?
    if (null == p_oExistingArray) {
      Object[][] p_oReturnArray = new Object[1][];
      p_oReturnArray[0] = p_oAddArray;
      return p_oReturnArray;

    }
    else {

      //Is there space in the existing array?
      int i;
      for (i = 0; i < p_oExistingArray.length; i++) {
        if (p_oExistingArray[i] == null) {
          p_oExistingArray[i] = p_oAddArray;
          return p_oExistingArray;
        }
      }

      Object[][] p_oReturnArray = new Object[p_oExistingArray.length + 1][];
      for (i = 0; i < p_oExistingArray.length; i++) {
        p_oReturnArray[i] = p_oExistingArray[i];
      }

      p_oReturnArray[p_oExistingArray.length] = p_oAddArray;
      return p_oReturnArray;
    }
  }

  /**
   * Creates an object array of strings consisting of species names, suitable
   * for including in the parameter entry table.  The species will be in order,
   * skipping over any positions in p_bUsed which are false.  The first
   * position in the array is blank.
   * @param p_bUsed Array of which species to include, in species number order.
   * @param oPop TreePopulation object.
   * @return Object array of species names, or null if no species are to be
   * used.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected Object[] FormatSpeciesHeaderRow(boolean[] p_bUsed,
                                            TreePopulation oPop) {
    Object[] p_oSpecies;
    int iNumTotalSpecies = oPop.GetNumberOfSpecies(),
        iNumUsedSpecies = 0, i;

    //Count how many species are used
    for (i = 0; i < iNumTotalSpecies; i++) {
      if (p_bUsed[i]) {
        iNumUsedSpecies++;
      }
    }

    if (iNumUsedSpecies == 0) {
      return null;
    }
    p_oSpecies = new Object[iNumUsedSpecies + 1];
    p_oSpecies[0] = " ";
    iNumUsedSpecies = 1;
    for (i = 0; i < iNumTotalSpecies; i++) {
      if (p_bUsed[i]) {
        p_oSpecies[iNumUsedSpecies] = oPop.GetSpeciesNameFromCode(i).replace(
            '_', ' ');
        iNumUsedSpecies++;
      }
    }

    return p_oSpecies;
  }

  /**
   * Creates a blank header row for a table.  It will be sized the same size
   * as the size of the array in the first array position.
   * @param p_oTable The table for which to create the blank header row.
   * @return The blank header row.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected Object[] FormatBlankHeaderRow(Object[][] p_oTable) {
    Object[] p_oHeader = new Object[p_oTable[0].length];
    int i;

    for (i = 0; i < p_oHeader.length; i++) {
      p_oHeader[i] = " "; //a space ensures the header row has the right
      //height
    }

    return p_oHeader;
  }

  /**
   * Wraps up a header row and table data into a vector.
   * @param p_oHeader Header row.
   * @param p_oData Table data.
   * @return Table vector.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public static Vector MakeTableVector(Object[] p_oHeader, Object[][] p_oData) {
    Vector p_oTableVector = new Vector(2);
    p_oTableVector.add(p_oHeader);
    p_oTableVector.add(p_oData);
    return p_oTableVector;
  }

  /**
   * Adds a value with combo box formatting to an Object array for display
   * in a JTable.  The formatting consists of some special characters in a
   * text string that our parameter display code will recognize as setup for
   * a combo box.
   * @param p_oExisting Object array to add to.
   * @param oToFormat Value to add.
   * @return Object array with new value added.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected Object[][] FormatDataAsComboBox(Object[][] p_oExisting,
                                            ModelEnum oToFormat) {
    if (null == oToFormat) {
      return p_oExisting;
    }

    //Create our array
    Object[] p_oArray = new Object[2];

    p_oArray[0] = oToFormat.GetDescriptor();
    p_oArray[1] = FormatComboBoxString(oToFormat);

    return AddDataToArray(p_oExisting, p_oArray);

  }

  /**
   * Formats a combo box string from a ModelEnum.  The formatting consists of
   * some special characters in a text string that our parameter display code
   * will recognize as setup for a combo box.
   * @param oToFormat ModelEnum to format.
   * @return Formatted string.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected String FormatComboBoxString(ModelEnum oToFormat) {
    String[] p_sLabels = oToFormat.GetAllowedValueLabels();
    String sChoices = "&&" + oToFormat.GetStringValue() + "|" +
        p_sLabels[0];
    int i;
    for (i = 1; i < p_sLabels.length; i++) {
      sChoices = sChoices + "," + p_sLabels[i];
    }
    return sChoices;
  }

  /**
   * Reads in data from the parameter display window.  The format of the
   * incoming data is the same as that which originally went jOut to be
   * displayed in a table - a vector of vectors, with each child vector
   * representing one table with an Object[] array header as the first element
   * and an Object[][] array of data as the second.
   * @param p_oData Data to read in.
   * @param oPop TreePopulation object.
   * @throws ModelException if the data cannot be read into the object.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void ReadDataFromDisplay(Vector p_oData, TreePopulation oPop) throws
      ModelException {
    Vector p_oOneTable;
    Object[][] p_oOneTableData;
    Object[] p_oHeader;
    boolean[] p_bAppliesTo;
    String sVal, sDataName = null;
    int i, j, k, iRow,
        iNumSpecies = oPop.GetNumberOfSpecies();
    boolean bIsSingle;

    p_bAppliesTo = new boolean[iNumSpecies];

    //There are an unknown number of tables - process them one at a time
    for (i = 0; i < p_oData.size(); i++) {

      p_oOneTable = (Vector) p_oData.elementAt(i);

      //Split into header and table data
      p_oHeader = (Object[]) p_oOneTable.elementAt(0);
      p_oOneTableData = (Object[][]) p_oOneTable.elementAt(1);

      //Is the header a row of species?  Check the last array bucket to see if
      //it's got a value
      sVal = (String) p_oHeader[p_oHeader.length - 1];
      if (sVal.trim().length() > 0) {

        bIsSingle = false;

        //Yes - should be species - transform to an array of booleans for
        //which species are present
        for (j = 0; j < p_bAppliesTo.length; j++) {
          p_bAppliesTo[j] = false;
        }
        for (j = 1; j < p_oHeader.length; j++) {
          sVal = (String) p_oHeader[j];
          sVal = sVal.replace(' ', '_');
          p_bAppliesTo[oPop.GetSpeciesCodeFromName(sVal)] = true;
        }

      }
      else {
        bIsSingle = true;
      }

      //Go through the data rows one at a time
      for (iRow = 0; iRow < p_oOneTableData.length; iRow++) {

        //Get the string data name
        sDataName = (String) p_oOneTableData[iRow][0];

        //Find the object
        if (sDataName != null && sDataName.trim().length() > 0) {
          ModelData oDataMember = FindObjectByDescriptor(sDataName);
          if (oDataMember == null) {
            ModelException oErr = new ModelException();
            oErr.sMessage = "Couldn't find a place in " + m_sName +
                " to put "
                + sDataName + ".";
            oErr.sCallingFunction = "JAVA";
            throw (oErr);
          }

          //Figure jOut what to do based on data type
          //ModelVector object
          if (oDataMember instanceof ModelVector) {

            ModelVector p_oVectorData = (ModelVector) oDataMember;

            //If not species-specific, panic
            if (bIsSingle) {
              ModelException oErr = new ModelException();
              oErr.sMessage = "Bad data passed to " + m_sName +
                  " for the value "
                  + sVal;
              oErr.sCallingFunction = "JAVA";
              throw (oErr);
            }

            //Make the table row into a vector of strings sized to the
            //number of species
            Vector p_oValuesToSet = new Vector(iNumSpecies);
            k = 1;
            for (j = 0; j < iNumSpecies; j++) {
              if (p_bAppliesTo[j]) {
                sVal = EnhancedTable.GetComboValue(String.valueOf(
                    p_oOneTableData[iRow][k]));
                if (sVal.equals("null")) {
                  p_oValuesToSet.add(null);
                }
                else {
                  p_oValuesToSet.add(sVal);
                }
                k++;
              }
              else {
                p_oValuesToSet.add(null);
              }
            }
            SetVectorValues(p_oVectorData, p_oValuesToSet, p_bAppliesTo);
          }

          //Assume it's a single value
          else {
            SetSingleValue(oDataMember,
                           EnhancedTable.GetComboValue(String.valueOf(
                p_oOneTableData[
                iRow][1])));
          }
        }
      }
    }
  }

  /**
   * Override this to implement data checking.  This will be called after
   * data has been set.
   * @param oPop TreePopulation object.
   * @throws ModelException if data doesn't pass validation.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  abstract public void ValidateData(TreePopulation oPop) throws ModelException;

  /**
   * Adds a data object to the arrays which will eventually display run
   * parameters in a set of tables.  There are two possible vectors to which
   * to add the data object - one grouping single values, and other grouping
   * groups of species-specific values according to the species they require.
   * <p>
   * When a new data object is to be added to the list, the first thing to do
   * is see if it's already there.  If it is, and it's not species-specific,
   * it's not added again.  If it is species-specific, the species to which it
   * applies are merged with the existing list and it is re-grouped if
   * necessary.
   * @param p_oDataObject Data object to add
   * @param p_oSingles Set of single values
   * @param p_oSpeciesSpecific Set of species-specific values
   * @param p_bAppliesTo If this is a species-specific value, what it
   * applies to
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected void AddDataObjectToDisplayArrays(ModelData p_oDataObject,
                                              Vector p_oSingles,
                                              Vector p_oSpeciesSpecific,
                                              boolean[] p_bAppliesTo) {

    Vector p_oSpecSpecGroup;
    SpeciesSpecific p_oExisting, p_oNew;
    ModelData p_oExistingDataObject;
    //This is the maximum number of rows to put in a species-specific table.
    //When we hit this number we'll start a new table.  This way the headers
    //never get too far away from the parameters.
    int iMaxRows = 10, i, j;
    boolean bIsDuplicate, bPlaceExistsForThis;
    //Set the flag as to whether this is a duplicate to false
    bIsDuplicate = false;

    //Process this data object according to what kind it is
    //Single data objects
    if (p_oDataObject instanceof ModelInt ||
        p_oDataObject instanceof ModelFloat ||
        p_oDataObject instanceof ModelEnum ||
        p_oDataObject instanceof ModelString) {

      //See if there's a duplicate in the singles list
      for (i = 0; i < p_oSingles.size(); i++) {
        p_oExistingDataObject = (ModelData) p_oSingles.elementAt(i);
        if (p_oExistingDataObject.GetDescriptor().equals(p_oDataObject.
            GetDescriptor())) {
          bIsDuplicate = true;
        }
      }

      //Not a duplicate?  Add it
      if (!bIsDuplicate) {
        p_oSingles.add(p_oDataObject);
      }
    }
    else if (p_oDataObject instanceof ModelVector) {

      p_oNew = new SpeciesSpecific();
      p_oNew.m_oData = p_oDataObject;
      p_oNew.mp_bUsed = p_bAppliesTo;

      //Is this the first one?
      if (p_oSpeciesSpecific.size() == 0) {
        p_oSpecSpecGroup = new Vector(1);
        p_oSpecSpecGroup.add(p_oNew);
        p_oSpeciesSpecific.add(p_oSpecSpecGroup);
      }
      else {

        //This is a model vector - see if it's a duplicate
        for (i = 0; i < p_oSpeciesSpecific.size(); i++) {
          p_oSpecSpecGroup = (Vector) p_oSpeciesSpecific.elementAt(i);
          for (j = 0; j < p_oSpecSpecGroup.size(); j++) {
            p_oExisting = (SpeciesSpecific) p_oSpecSpecGroup.elementAt(j);
            p_oExistingDataObject = (ModelData) p_oExisting.m_oData;
            if (p_oExistingDataObject.GetDescriptor().equals(p_oDataObject.
                GetDescriptor())) {
              bIsDuplicate = true;

              //Does it have the same species list?  If not, merge them,
              //remove the existing object, and let the new one get
              //treated as a new object
              if (!BoolArraysEqual(p_oExisting.mp_bUsed, p_oNew.mp_bUsed)) {
                bIsDuplicate = false;
                p_oSpecSpecGroup.remove(j);
                if (p_oSpecSpecGroup.size() == 0) {
                  p_oSpeciesSpecific.remove(i);
                }
                p_oNew.mp_bUsed = MergeBooleans(p_oNew.mp_bUsed,
                                                p_oExisting.mp_bUsed);
                break;
              }
            }
          }
        }

        if (!bIsDuplicate) {
          //Go through again and see if there is an existing species-specific
          //group which applies to the same species
          bPlaceExistsForThis = false;
          for (i = 0; i < p_oSpeciesSpecific.size(); i++) {
            p_oSpecSpecGroup = (Vector) p_oSpeciesSpecific.elementAt(i);
            p_oExisting = (SpeciesSpecific) p_oSpecSpecGroup.elementAt(0);

            if (BoolArraysEqual(p_oExisting.mp_bUsed, p_oNew.mp_bUsed)) {
              //Make sure the existing table doesn't have too many rows
              if (p_oSpecSpecGroup.size() < iMaxRows) {
                bPlaceExistsForThis = true;
                p_oSpecSpecGroup.add(p_oNew);
              }
            }
          }

          if (!bPlaceExistsForThis) {
            //Create a new species-specific group and put it in
            p_oSpecSpecGroup = new Vector(1);
            p_oSpecSpecGroup.add(p_oNew);
            p_oSpeciesSpecific.add(p_oSpecSpecGroup);
          }
        }
      }
    }
    else {
      throw (new java.lang.RuntimeException(
          "FormatDataForDisplay can't deal with this."));
    }

  }

  /**
   * Merges two boolean arrays and returns an array where there is a true
   * in each array bucket where EITHER array is true.  If either array is null,
   * the other array is returned.  This will not make sure the arrays are the
   * same length.
   * @param p_oMergeOne First array to merge.
   * @param p_oMergeTwo Second array to merge.
   * @return Merged array.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected boolean[] MergeBooleans(boolean[] p_oMergeOne,
                                    boolean[] p_oMergeTwo) {
    if (p_oMergeOne == null) {
      return p_oMergeTwo;
    }
    if (p_oMergeTwo == null) {
      return p_oMergeOne;
    }

    boolean[] p_oMerged = new boolean[p_oMergeOne.length];
    int i;
    for (i = 0; i < p_oMerged.length; i++) {
      if (p_oMergeOne[i] || p_oMergeTwo[i]) {
        p_oMerged[i] = true;
      }
      else {
        p_oMerged[i] = false;
      }
    }

    return p_oMerged;
  }

  /**
   * Tests two boolean arrays for equality.  Equality is defined as having the
   * same value in each array bucket.
   * @param p_bArrayOne First array.
   * @param p_bArrayTwo Second array.
   * @return true if the arrays are equal, false if they are not.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected boolean BoolArraysEqual(boolean[] p_bArrayOne,
                                    boolean[] p_bArrayTwo) {
    if (p_bArrayOne == null && p_bArrayTwo == null) {
      return true;
    }
    if ( (p_bArrayOne == null && p_bArrayTwo != null) ||
        p_bArrayOne != null && p_bArrayTwo == null) {
      return false;
    }
    if (p_bArrayOne.length != p_bArrayTwo.length) {
      return false;
    }

    int i;
    for (i = 0; i < p_bArrayOne.length; i++) {
      if (p_bArrayOne[i] != p_bArrayTwo[i]) {
        return false;
      }
    }
    return true;
  }

  /**
   * Turns vector groups of data objects into a set of tables for display.
   * @param p_oSingles Single values.
   * @param p_oSpeciesSpecific Vector of vectors of species-specific values
   * grouped together by what species they apply to.
   * @param oPop TreePopulation object.
   * @return Vector of tables for display.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected Vector FormatTable(Vector p_oSingles, Vector p_oSpeciesSpecific,
                               TreePopulation oPop) {
    Object[][] p_oTable = null;
    Vector p_oSpecSpecGroup,
        p_oWhatToReturn = new Vector();
    SpeciesSpecific p_oExisting;
    boolean[] p_bAppliesTo = null;
    int i, j;

    //Format what we got into a table
    if (p_oSingles.size() == 0 && p_oSpeciesSpecific.size() == 0) {
      return null;
    }

    p_oTable = null;
    for (i = 0; i < p_oSingles.size(); i++) {
      if (p_oSingles.elementAt(i) instanceof ModelInt) {
        ModelInt oData = (ModelInt) p_oSingles.elementAt(i);
        p_oTable = FormatDataForTable(p_oTable, oData);
      }
      else if (p_oSingles.elementAt(i) instanceof ModelFloat) {
        ModelFloat oData = (ModelFloat) p_oSingles.elementAt(i);
        p_oTable = FormatDataForTable(p_oTable, oData);
      }
      else if (p_oSingles.elementAt(i) instanceof ModelEnum) {
        ModelEnum oData = (ModelEnum) p_oSingles.elementAt(i);
        p_oTable = FormatDataAsComboBox(p_oTable, oData);
      }
      else if (p_oSingles.elementAt(i) instanceof ModelString) {
        ModelString oData = (ModelString) p_oSingles.elementAt(i);
        p_oTable = FormatDataForTable(p_oTable, oData);
      }

    }
    if (p_oTable != null) {
      p_oWhatToReturn.add(MakeTableVector(FormatBlankHeaderRow(p_oTable),
                                          p_oTable));
    }

    for (i = 0; i < p_oSpeciesSpecific.size(); i++) {
      p_oSpecSpecGroup = (Vector) p_oSpeciesSpecific.elementAt(i);
      p_oTable = null;
      for (j = 0; j < p_oSpecSpecGroup.size(); j++) {
        p_oExisting = (SpeciesSpecific) p_oSpecSpecGroup.elementAt(j);
        p_oTable = FormatDataForTable(p_oTable,
                                      (ModelVector) p_oExisting.m_oData,
                                      p_oExisting.mp_bUsed);
        p_bAppliesTo = p_oExisting.mp_bUsed;
      }
      if (p_oTable != null) {
        p_oWhatToReturn.add(MakeTableVector(FormatSpeciesHeaderRow(p_bAppliesTo,
            oPop), p_oTable));
      }
    }
    return p_oWhatToReturn;
  }

  /**
   * Finds an object based on its XML tag.
   * @param sXMLTag XML tag for the object.
   * @return ModelData object corresponding to the XML tag, or null if no such
   * object is found.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public ModelData FindObjectByXMLTag(String sXMLTag) {
    ModelData oDataMember;
    int i;
    for (i = 0; i < mp_oAllData.size(); i++) {
      oDataMember = (ModelData) mp_oAllData.elementAt(i);
      if (oDataMember.GetXMLTag().equals(sXMLTag)) {
        return oDataMember;
      }
    }
    return null;
  }

  /**
   * Finds an object based on its descriptor.
   * @param sDescriptor Descriptor for the object.
   * @return ModelData object corresponding to the descriptor, or null if
   * no such object is found.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public ModelData FindObjectByDescriptor(String sDescriptor) {
    ModelData oDataMember;
    int i;
    for (i = 0; i < mp_oAllData.size(); i++) {
      oDataMember = (ModelData) mp_oAllData.elementAt(i);
      if (oDataMember.GetDescriptor().equals(sDescriptor)) {
        return oDataMember;
      }
    }
    return null;
  }

  /**
   * Sets a data object's value.  Override this to add functionality.
   * @param sXMLTag XML tag of data object whose value is to be set.
   * @param sXMLParentTag The immediate parent tag that sXMLTag is within.
   * @param oAttributes Attributes of the object.  Ignored, but may be needed
   * by overriding objects.
   * @param oData Data value, either a String or a type appropriate to the
   * data type
   * @return true if the value was set successfully; false if the value could
   * not be found.  (This would not be an error, because I need a way to
   * cycle through the objects until one of the objects comes up with a
   * match.)
   * @throws ModelException if the value could not be assigned to the data
   * object.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean SetSingleValueByXMLTag(String sXMLTag, String sXMLParentTag,
                                        Attributes oAttributes,
                                        Object oData) throws
      ModelException {
    ModelData oDataMember = FindObjectByXMLTag(sXMLTag);
    if (oDataMember == null) {
      return false;
    }

    SetSingleValue(oDataMember, oData);
    return true;
  }

  /**
   * Sets the value of a single data member.  The value to set is cast to the
   * proper type.
   * @param oDataMember Data member to set.
   * @param oData Data value to set into data member.
   * @throws ModelException if data is of the wrong type.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected void SetSingleValue(ModelData oDataMember, Object oData) throws
      ModelException {
    try {
      if (oDataMember instanceof ModelFloat) {
        ModelFloat oFloatDataMember = (ModelFloat) oDataMember;
        Float oFloat;
        if (oData instanceof String) {
          oFloat = new Float( (String) oData);
        }
        else {
          oFloat = (Float) oData;
        }
        oFloatDataMember.SetValue(oFloat.floatValue());
      }
      else if (oDataMember instanceof ModelInt) {
        ModelInt oIntDataMember = (ModelInt) oDataMember;
        Integer oInt;
        if (oData instanceof String) {
          oInt = new Integer( (String) oData);
        }
        else {
          oInt = (Integer) oData;
        }
        oIntDataMember.SetValue(oInt.intValue());
      }
      else if (oDataMember instanceof ModelEnum) {
        ModelEnum oEnumDataMember = (ModelEnum) oDataMember;
        Integer oInt;
        if (oData instanceof String) {
          //Try to make the string into an integer - if it doesn't work,
          //set the string directly
          try {
            oInt = new Integer( (String) oData);
            oEnumDataMember.SetValue(oInt.intValue());
          }
          catch (java.lang.NumberFormatException oE) {
            oEnumDataMember.SetValue( (String) oData);
          }
        }
        else {
          oInt = (Integer) oData;
          oEnumDataMember.SetValue(oInt.intValue());
        }
      }
      else if (oDataMember instanceof ModelString) {
        ModelString oStringDataMember = (ModelString) oDataMember;
        oStringDataMember.SetValue( (String) oData);
      }
      else {
        //Panic
        ModelException oErr = new ModelException();
        oErr.sMessage = "For object " + m_sName + ", data " +
            oDataMember.GetDescriptor() + ", can't determine data type.";
        oErr.sCallingFunction = "JAVA";
        throw (oErr);
      }
    }
    catch (java.lang.ClassCastException oErr) {
      ModelException oExp = new ModelException();
      if (oDataMember != null) {
        oExp.sMessage = "Attempt to set " + oDataMember.GetDescriptor() +
            " with the wrong data type.";
      }
      oExp.sCallingFunction = "JAVA";
      throw (oExp);
    }
    catch (java.lang.NumberFormatException oErr) {
      ModelException oExp = new ModelException();
      if (oDataMember != null) {
        oExp.sMessage = "Attempt to set " + oDataMember.GetDescriptor() +
            " with the wrong data type.";
      }
      oExp.sCallingFunction = "JAVA";
      throw (oExp);
    }
  }

  /**
   * Sets a data vector's value.  Override this to add functionality.
   * @param sXMLTag Parent XML tag of data vector whose value is to be set.
   * @param sXMLParentTag The immediate parent tag that sXMLTag is within.
   * @param p_oData Vector of data values appropriate to the data type
   * @param p_sChildXMLTags The XML tags of the child elements
   * @param p_bAppliesTo Array of booleans saying which of the vector values
   * should be set.  This is important in the case of species-specifics - the
   * vector index is the species number but not all species are set.
   * @param oParentAttributes Attributes of parent tag.  May be useful when
   * overridding this for unusual tags.
   * @param p_oAttributes Attributes passed from parser.  This may be needed
   * when overriding this function.  Basic species-specific values are
   * already handled by this function.
   * @return true if the value was set successfully; false if the value could
   * not be found.  (This would not be an error, because I need a way to
   * cycle through the objects until one of the objects comes up with a
   * match.)  If a match to a data object is made via XML tag, but the found
   * object is not a ModelVector, this returns false.
   * @throws ModelException if the value could not be assigned to the data
   * object.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean SetVectorValueByXMLTag(String sXMLTag, String sXMLParentTag,
                                        Vector p_oData,
                                        String[] p_sChildXMLTags,
                                        boolean[] p_bAppliesTo,
                                        Attributes oParentAttributes,
                                        Attributes[] p_oAttributes) throws
      ModelException {
    if (sXMLTag == null || sXMLTag.length() == 0) {
      return false;
    }

    ModelData oDataMember = FindObjectByXMLTag(sXMLTag);
    if (oDataMember == null) {
      return false;
    }
    if (! (oDataMember instanceof ModelVector)) {
      return false;
    }
    if (p_bAppliesTo.length < p_oData.size()) {
      ModelException oErr = new ModelException();
      oErr.sMessage = "For " + m_sName +
          " - trying to assign an array that is " +
          "too large to " + oDataMember.GetDescriptor();
      throw (oErr);
    }
    SetVectorValues( (ModelVector) oDataMember, p_oData, p_bAppliesTo);

    return true;
  }

  /**
   * Accepts an XML parent tag (empty, no data) from the parser.  Normally this
   * data would be ignored, but this can be overridden if there is special
   * XML to be captured.
   * @param sXMLTag The XML tag.
   * @param oAttributes The attributes of this object.
   * @throws ModelException if there is a problem reading this data.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void ReadXMLParentTag(String sXMLTag, Attributes oAttributes) throws
      ModelException {}

  /**
   * Triggered when there is a change in the species list.  This goes through
   * everything in the mp_oAllData array.  Any piece of data of type
   * ModelVector which has a size matching the old number of species is
   * assumed to be species-specific and in need of transformation.  The
   * ModelVector will be re-sized to match the new number of species.  Any
   * species that are the same from the old set to the new have their data
   * transferred to their new array index.  Any new species indexes are set
   * to null.  Any species which were deleted have their data lost.
   * @param iOldNumSpecies says how many species there used to be.
   * @param p_iIndexer is an array, sized to the new number of species.  For
   * each bucket (representing the index number of a species on the new list),
   * the value is either the index of that same species in the old species
   * list, or -1 if the species is new.
   * @throws ModelException if there's a problem.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void ChangeOfSpecies(int iOldNumSpecies, int[] p_iIndexer) throws ModelException {
    ModelVector p_oVectorData; //A ModelVector object that needs translation
    Vector p_oCopy = new Vector(iOldNumSpecies); //A copy of the old data
    int i, j;

    for (i = 0; i < mp_oAllData.size(); i++) {
      if (mp_oAllData.elementAt(i) instanceof ModelVector) {
        p_oVectorData = (ModelVector) mp_oAllData.elementAt(i);

        //Is it sized number of species?
        if (p_oVectorData.GetValue().size() == iOldNumSpecies) {

          //Make a copy of the vector's data
          for (j = 0; j < iOldNumSpecies; j++) {
            p_oCopy.add(j, p_oVectorData.GetValue().elementAt(j));
          }

          //Clear jOut the existing array
          p_oVectorData.GetValue().clear();

          //Now transfer back any data that has the same species
          for (j = 0; j < p_iIndexer.length; j++) {
            if (p_iIndexer[j] != -1) {
              p_oVectorData.GetValue().add(p_oCopy.elementAt(p_iIndexer[j]));
            }
            else {
              p_oVectorData.GetValue().add(null);
            }
          }
        }
      }
    }
  }

  /**
   * Performs any necessary tasks associated with changing the name of a
   * species.  Everything about the species remains the same except for the
   * name.  This is an easier process than actually changing the species list.
   * @param sOldSpecies String Old name of the species, with underscores
   * instead of spaces (like the species names would come from the tree
   * population)
   * @param sNewSpecies String New name of the species, with underscores
   * instead of spaces (like the species names would come from the tree
   * population)
   * @throws ModelException if there is a problem.
   */
  public void ChangeOfSpeciesName (String sOldSpecies, String sNewSpecies) throws ModelException {}
}

/**This class will package together a data member and an array of booleans
    on which species uses it.
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 */
class SpeciesSpecific {
  public Object m_oData;
  public boolean[] mp_bUsed;
}
